from typing import Any

import pytest

from tests.opentrons.conftest import Protocol

from opentrons.protocol_api import ParameterContext
from opentrons.protocol_api.protocol_context import ProtocolContext
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.execution import execute, execute_python
from opentrons.protocols.execution.errors import ExceptionInProtocolError
from opentrons.protocols.parse import parse
from opentrons.protocols.types import MalformedPythonProtocolError, PythonProtocol


def test_api2_runfunc() -> None:
    def noargs() -> None:
        pass

    with pytest.raises(SyntaxError):
        execute_python._runfunc_ok(noargs)

    def twoargs(a: Any, b: Any) -> None:
        pass

    with pytest.raises(SyntaxError):
        execute_python._runfunc_ok(twoargs)

    def two_with_default(a: Any, b: Any = 2) -> None:
        pass

    # making sure this doesn't raise
    execute_python._runfunc_ok(two_with_default)

    def one_with_default(a: Any = 2) -> None:
        pass

    # shouldn't raise
    execute_python._runfunc_ok(one_with_default)

    def starargs(*args: Any) -> None:
        pass

    # shouldn't raise
    execute_python._runfunc_ok(starargs)


@pytest.mark.ot2_only
@pytest.mark.parametrize("protocol_file", ["testosaur_v2.py"])
def test_execute_ok(
    protocol: Protocol, protocol_file: str, ctx: ProtocolContext
) -> None:
    proto = parse(protocol.text, protocol.filename)
    execute.run_protocol(proto, context=ctx)


def test_bad_protocol(ctx: ProtocolContext) -> None:
    no_args = parse(
        """
metadata={"apiLevel": "2.0"}
def run():
    pass
"""
    )
    with pytest.raises(MalformedPythonProtocolError) as e:
        execute.run_protocol(no_args, context=ctx)
        assert "Function 'run()' does not take any parameters" in str(e.value)

    many_args = parse(
        """
metadata={"apiLevel": "2.0"}
def run(a, b):
    pass
"""
    )
    with pytest.raises(MalformedPythonProtocolError) as e:
        execute.run_protocol(many_args, context=ctx)
        assert "must be called with more than one argument" in str(e.value)


def test_proto_with_exception(ctx: ProtocolContext) -> None:
    exc_in_root = """metadata={"apiLevel": "2.0"}

def run(ctx):
    raise Exception("hi")
"""
    protocol = parse(exc_in_root)
    with pytest.raises(ExceptionInProtocolError) as e:
        execute.run_protocol(protocol, context=ctx)
    assert "Exception [line 4]: hi" in str(e.value)

    nested_exc = """
import ast

def this_throws():
    raise Exception("hi")

def run(ctx):
    this_throws()

metadata={"apiLevel": "2.0"};
"""
    protocol = parse(nested_exc)
    with pytest.raises(ExceptionInProtocolError) as e:
        execute.run_protocol(protocol, context=ctx)
    assert "[line 5]" in str(e.value)
    assert "Exception [line 5]: hi" in str(e.value)


# TODO (spp, 2024-7-16): add a test for CSV rtp extraction
@pytest.mark.ot2_only
@pytest.mark.parametrize("protocol_file", ["testosaur_with_rtp.py"])
def test_rtp_extraction(protocol: Protocol, protocol_file: str) -> None:
    """It should set the RTP definitions in protocol with override values from run."""
    proto = parse(protocol.text, protocol.filename)
    parameter_context = ParameterContext(api_version=APIVersion(2, 18))
    run_time_param_overrides = {"sample_count": 2}
    assert isinstance(proto, PythonProtocol)
    val = execute_python.exec_add_parameters(
        protocol=proto,
        parameter_context=parameter_context,
        run_time_param_overrides=run_time_param_overrides,
        run_time_param_file_overrides={},
    )
    assert val is not None
    assert val.get_all() == {"sample_count": 2, "mount": "left"}



"""Fixtures for protocol api integration tests."""

from typing import Generator

import pytest
from _pytest.fixtures import SubRequest

from opentrons import protocol_api, simulate
from opentrons.protocol_api.core.engine import ENGINE_CORE_API_VERSION


@pytest.fixture
def simulated_protocol_context(
    request: SubRequest,
) -> Generator[protocol_api.ProtocolContext, None, None]:
    """Return a protocol context with requested version and robot."""
    version, robot_type = request.param
    context = simulate.get_protocol_api(version=version, robot_type=robot_type)
    try:
        yield context
    finally:
        if context.api_version >= ENGINE_CORE_API_VERSION:
            # TODO(jbl, 2024-11-14) this is a hack of a hack to close the hardware and the PE thread when a test is
            #  complete. At some point this should be replaced with a more holistic way of safely cleaning up these
            #  threads so they don't leak and cause tests to fail when `get_protocol_api` is called too many times.
            simulate._LIVE_PROTOCOL_ENGINE_CONTEXTS.close()
        else:
            # If this is a non-PE context we need to clean up the hardware thread manually
            context._hw_manager.hardware.clean_up()



from copy import deepcopy
from typing import Dict, cast

import pytest
from mock import AsyncMock, patch

from opentrons.drivers import utils
from opentrons.drivers.asyncio.communication import AlarmResponse
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.rpi_drivers.dev_types import GPIODriverLike
from opentrons.drivers.rpi_drivers.gpio_simulator import SimulatingGPIOCharDev
from opentrons.drivers.smoothie_drivers import constants, driver_3_0
from opentrons.drivers.smoothie_drivers.connection import SmoothieConnection
from opentrons.drivers.smoothie_drivers.constants import (
    HOMED_POSITION,
    Y_BOUND_OVERRIDE,
)
from opentrons.drivers.smoothie_drivers.errors import SmoothieAlarm, SmoothieError


@pytest.fixture
async def mock_connection() -> AsyncMock:
    """The mock SerialConnection."""
    return AsyncMock(spec=SmoothieConnection)


@pytest.fixture
def sim_gpio() -> SimulatingGPIOCharDev:
    """The mock GPIO."""
    return SimulatingGPIOCharDev("sim")


@pytest.fixture
def smoothie(
    mock_connection: AsyncMock, sim_gpio: GPIODriverLike
) -> driver_3_0.SmoothieDriver:
    """The smoothie driver under test."""
    from opentrons.config import robot_configs

    d = driver_3_0.SmoothieDriver(
        connection=mock_connection,
        config=robot_configs.load_ot2(),
        gpio_chardev=sim_gpio,
    )
    return d


def position(
    x: float, y: float, z: float, a: float, b: float, c: float
) -> Dict[str, float]:
    return {axis: value for axis, value in zip("XYZABC", [x, y, z, a, b, c])}


async def test_update_position(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """It should update the position."""
    mock_connection.send_command.return_value = (
        "ok MCS: X:1.0000 Y:2.0000 Z:3.0000 A:4.5000 B:0.0000 C:0.0000"
    )

    await smoothie.update_position()
    expected = {"X": 1, "Y": 2, "Z": 3, "A": 4.5, "B": 0, "C": 0}
    assert smoothie.position == expected


async def test_update_position_retry(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """It should retry after a parse error."""
    mock_connection.send_command.side_effect = [
        # first attempt to read, we get bad data
        "ok MCS: X:0.0000 Y:MISTAKE Z:0.0000 A:0.0000 B:0.0000 C:0.0000",
        "ok",
        # following attempts to read, we get good data
        "ok MCS: X:0.0000 Y:321.00 Z:0.0000 A:0.0000 B:0.0000 C:0.0000",
        "ok",
    ]

    await smoothie.update_position()
    expected = {"X": 0, "Y": 321.00, "Z": 0, "A": 0, "B": 0, "C": 0}
    assert smoothie.position == expected


def test_active_dwelling_current_push_pop(smoothie: driver_3_0.SmoothieDriver) -> None:
    assert smoothie._active_current_settings != smoothie._dwelling_current_settings

    old_active_currents = deepcopy(smoothie._active_current_settings)
    old_dwelling_currents = deepcopy(smoothie._dwelling_current_settings)

    smoothie.push_active_current()
    smoothie.set_active_current({"X": 2.0, "Y": 2.0, "Z": 2.0, "A": 2.0})
    smoothie.pop_active_current()

    assert smoothie._active_current_settings == old_active_currents
    assert smoothie._dwelling_current_settings == old_dwelling_currents


async def test_functional(smoothie: driver_3_0.SmoothieDriver) -> None:
    smoothie.simulating = True
    assert smoothie.position == position(0, 0, 0, 0, 0, 0)

    await smoothie.move({"X": 0, "Y": 1, "Z": 2, "A": 3, "B": 4, "C": 5})
    assert smoothie.position == position(0, 1, 2, 3, 4, 5)

    await smoothie.move({"X": 1, "Z": 3, "C": 6})
    assert smoothie.position == position(1, 1, 3, 3, 4, 6)

    await smoothie.home(axis="abc", disabled="")
    assert smoothie.position == position(
        1,
        1,
        3,
        smoothie.homed_position["A"],
        smoothie.homed_position["B"],
        smoothie.homed_position["C"],
    )

    await smoothie.home(disabled="")
    assert smoothie.position == smoothie.homed_position


async def test_read_pipette_v13(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    mock_connection.send_command.return_value = "L:" + utils.string_to_hex(
        "p300_single_v13"
    )
    res = await smoothie.read_pipette_model("left")
    assert res == "p300_single_v1.3"


async def test_switch_state(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    smoothie_switch_res = (
        "X_max:0 Y_max:0 Z_max:0 A_max:0 B_max:0 C_max:0"
        " _pins "
        "(XL)2.01:0 (YL)2.01:0 (ZL)2.01:0 "
        "(AL)2.01:0 (BL)2.01:0 (CL)2.01:0 Probe: 0\r\n"
    )

    mock_connection.send_command.return_value = smoothie_switch_res

    expected = {
        "X": False,
        "Y": False,
        "Z": False,
        "A": False,
        "B": False,
        "C": False,
        "Probe": False,
    }
    r = await smoothie.switch_state()
    assert r == expected

    smoothie_switch_res = (
        "X_max:0 Y_max:0 Z_max:0 A_max:1 B_max:0 C_max:0"
        " _pins "
        "(XL)2.01:0 (YL)2.01:0 (ZL)2.01:0 "
        "(AL)2.01:0 (BL)2.01:0 (CL)2.01:0 Probe: 1\r\n"
    )

    mock_connection.send_command.return_value = smoothie_switch_res

    expected = {
        "X": False,
        "Y": False,
        "Z": False,
        "A": True,
        "B": False,
        "C": False,
        "Probe": True,
    }
    r = await smoothie.switch_state()
    assert r == expected


async def test_clear_limit_switch(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """
    This functions as a contract test around recovery from a limit-switch hit.
    Note that this *does not* itself guarantee correct physical behavior--this
    interaction has been designed and tested on the robot manually and then
    encoded in this test. If requirements change around physical behavior, then
    this test will need to be revised.
    """
    cmd_list = []

    async def write_mock(command: CommandBuilder, retries: int, timeout: float) -> str:
        cmd_list.append(command.build())
        if constants.GCODE.MOVE in command:
            raise AlarmResponse(port="", response="ALARM: Hard limit +C")
        elif constants.GCODE.CURRENT_POSITION in command:
            return "ok M114.2 X:10 Y:20 Z:30 A:40 B:50 C:60"
        elif constants.GCODE.HOMING_STATUS in command:
            return "X:1 Y:1 Z:1 A:1 B:1 C:1"
        else:
            return "ok"

    mock_connection.send_command.side_effect = write_mock

    # This will cause a limit-switch error and not back off
    with pytest.raises(SmoothieError):
        await smoothie.move({"C": 100})

    assert [c.strip() for c in cmd_list] == [
        # attempt to move and fail
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005 G0 C100.3 G0 C100",
        # recover from failure
        "M999",
        "M400",
        "G28.6",
        "M400",
        # set current for homing the failed axis (C)
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005 G28.2 C",
        "M400",
        # set current back to idling after home
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
        # update position
        "M114.2",
        "M400",
        "M907 A0.1 B0.05 C0.05 X0.3 Y0.3 Z0.1 G4 P0.005",
        "M400",
    ]


async def test_unstick_axes(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    cmd_list = []

    def write_mock(command: CommandBuilder, retries: int, timeout: float) -> str:
        cmd_list.append(command.build())
        if constants.GCODE.LIMIT_SWITCH_STATUS in command:
            return "X_max:0 Y_max:0 Z_max:0 A_max:0 B_max:0 C_max:0 Probe: 0 \r\n"
        else:
            return "ok"

    mock_connection.send_command.side_effect = write_mock

    await smoothie.unstick_axes(axes="ABCXYZ", distance=2, speed=3)

    assert [c.strip() for c in cmd_list] == [
        "M203.1 A3 B3 C3 X3 Y3 Z3",
        "M400",
        "M119",
        "M400",
        "M907 A0.5 B0.05 C0.05 X1.25 Y1.25 Z0.5 G4 P0.005 G0 A-2 B-2 C-2 X-2 Y-2 Z-2",
        "M400",
        "M907 A0.5 B0.05 C0.05 X1.25 Y1.25 Z0.5 G4 P0.005",
        "M400",
        "M203.1 A125 B40 C40 X600 Y400 Z125",
        "M400",
    ]


@pytest.mark.parametrize(
    argnames=["home_flags", "axis_string", "expected"],
    argvalues=[
        [{k: False for k in "XYZABC"}, "A", "A"],
        [{k: False for k in "XYZABC"}, "XA", "XA"],
        [{k: False for k in "XYZABC"}, "XY", "XY"],
        [{k: False for k in "XYZABC"}, "XYZABC", "XYZABC"],
        [
            {"A": True, "B": False, "C": True, "X": False, "Y": True, "Z": False},
            "XYZABC",
            "XZB",
        ],
    ],
)
async def test_home_flagged_axes(
    smoothie: driver_3_0.SmoothieDriver,
    home_flags: Dict[str, bool],
    axis_string: str,
    expected: str,
) -> None:
    """It should only home un-homed axes."""
    smoothie.home = AsyncMock()  # type: ignore[method-assign]
    await smoothie.update_homed_flags(home_flags)

    await smoothie.home_flagged_axes(axes_string=axis_string)

    smoothie.home.assert_called_once_with(expected)


@pytest.mark.parametrize(
    argnames=["home_flags", "axis_string"],
    argvalues=[
        [{k: True for k in "XYZABC"}, "A"],
        [{k: True for k in "XYZABC"}, "XYZABC"],
        [{"A": True, "B": False, "C": True, "X": False, "Y": True, "Z": False}, "ACY"],
    ],
)
async def test_home_flagged_axes_no_call(
    smoothie: driver_3_0.SmoothieDriver, home_flags: Dict[str, bool], axis_string: str
) -> None:
    """It should not home homed axes."""
    smoothie.home = AsyncMock()  # type: ignore[method-assign]
    await smoothie.update_homed_flags(home_flags)

    await smoothie.home_flagged_axes(axes_string=axis_string)

    smoothie.home.assert_not_called()


async def test_update_homed_flags_retry(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """It should retry."""
    mock_connection.is_open.return_value = True

    # Set all to false
    await smoothie.update_homed_flags({k: False for k in "ABCXYZ"})

    # First response is error then all axes homed
    mock_connection.send_command.side_effect = [
        # first attempt to read, we get bad data
        "ok AX:1 BBY:0 Z:0 A:0 B:0 C:0",
        "ok",
        "ok X:1 Y:1 Z:1 A:1 B:1 C:1",
        "ok",
    ]

    await smoothie.update_homed_flags()

    assert smoothie.homed_flags == {k: True for k in "ABCXYZ"}


def test_axis_bounds(smoothie: driver_3_0.SmoothieDriver) -> None:
    """It should override Y."""
    bounds = smoothie.axis_bounds

    assert bounds == {
        k: (v if k != "Y" else Y_BOUND_OVERRIDE) for k, v in HOMED_POSITION.items()
    }


async def test_set_max_speed(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """It should set the max speed."""
    await smoothie.set_axis_max_speed({"A": 22, "B": 322})
    cmds = [
        c.kwargs["command"].build().strip()
        for c in mock_connection.send_command.call_args_list
    ]
    assert cmds == ["M203.1 A22 B322", "M400"]


async def test_restore_axis_max_speed(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """It should set then restore the max speed."""
    smoothie._max_speed_settings = {"A": 1, "B": 2, "C": 3, "X": 4, "Y": 5, "Z": 6}
    async with smoothie.restore_axis_max_speed(
        {"A": 6, "B": 5, "C": 4, "X": 3, "Y": 2, "Z": 1}
    ):
        pass
    cmds = [
        c.kwargs["command"].build().strip()
        for c in mock_connection.send_command.call_args_list
    ]
    assert cmds == [
        # Set new max speed.
        "M203.1 A6 B5 C4 X3 Y2 Z1",
        "M400",
        # Restore old.
        "M203.1 A1 B2 C3 X4 Y5 Z6",
        "M400",
    ]


async def test_speed(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """It should set the speed."""
    await smoothie.set_speed(1)
    cmds = [
        c.kwargs["command"].build().strip()
        for c in mock_connection.send_command.call_args_list
    ]
    # 60 is (speed * seconds_per_minute)
    assert cmds == ["G0 F60", "M400"]


async def test_restore_speed(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """It should set then restore the speed."""
    smoothie._combined_speed = 3
    async with smoothie.restore_speed(2):
        pass
    cmds = [
        c.kwargs["command"].build().strip()
        for c in mock_connection.send_command.call_args_list
    ]
    assert cmds == [
        # Set speed.
        "G0 F120",
        "M400",
        # Restore old.
        "G0 F180",
        "M400",
    ]


async def test_alarm_handling_halt(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """Alarms should clear the halt flag and errors should not."""
    await smoothie.hard_halt()
    assert smoothie._is_hard_halting.is_set()

    with pytest.raises(SmoothieError):
        smoothie._handle_return("error: something else", is_error=True)

    assert smoothie._is_hard_halting.is_set()

    with pytest.raises(SmoothieAlarm):
        smoothie._handle_return("error: Alarm lock", is_alarm=True)

    assert not smoothie._is_hard_halting.is_set()


async def test_alarm_handling_nohalt(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """Alarms should be turned into errors outside of halt"""
    assert not smoothie._is_hard_halting.is_set()

    with pytest.raises(SmoothieError):
        smoothie._handle_return("Alarm: Hard limit -X", is_alarm=True)

    with pytest.raises(SmoothieError):
        smoothie._handle_return("error oh no!", is_error=True)


async def test_no_auto_recover_on_alarm(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """Alarms that dont get turned into errors should not incur recovery."""

    await smoothie.hard_halt()
    assert smoothie._is_hard_halting.is_set()

    with (
        patch.object(smoothie, "_send_command_unsynchronized"),
        patch.object(smoothie, "_reset_from_error"),
    ):
        mocked_send = cast(AsyncMock, smoothie._send_command_unsynchronized)
        mocked_send.side_effect = [SmoothieAlarm("G0", "Alarm: Hard limit -X")]
        with pytest.raises(SmoothieAlarm):
            await smoothie.move({"X": 10})
        mocked_send._reset_from_error.assert_not_called()


async def test_auto_recover_on_error(
    smoothie: driver_3_0.SmoothieDriver, mock_connection: AsyncMock
) -> None:
    """Errors do incur recovery."""
    with (
        patch.object(smoothie, "_send_command_unsynchronized"),
        patch.object(smoothie, "_reset_from_error"),
        patch.object(smoothie, "home"),
    ):
        mocked_home = cast(AsyncMock, smoothie.home)
        mocked_send = cast(AsyncMock, smoothie._send_command_unsynchronized)
        mocked_send.side_effect = [SmoothieError("G0", "error: Uh oh")]
        with pytest.raises(SmoothieError):
            await smoothie.move({"X": 10})
        mocked_send.assert_called_once()
        mocked_home.assert_called_once()



"""Test deck data provider."""

import pytest
from decoy import Decoy
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.resources import (
    DeckDataProvider,
    DeckFixedLabware,
    LabwareDataProvider,
)
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    DeckType,
)
from opentrons.types import DeckSlotName


@pytest.fixture
def mock_labware_data_provider(decoy: Decoy) -> LabwareDataProvider:
    """Get a mock in the shape of the LabwareDataProvider."""
    return decoy.mock(cls=LabwareDataProvider)


@pytest.mark.parametrize(
    ("deck_type", "expected_definition"),
    [
        (DeckType.OT2_STANDARD, lazy_fixture("ot2_standard_deck_def")),
        (DeckType.OT2_SHORT_TRASH, lazy_fixture("ot2_short_trash_deck_def")),
        (DeckType.OT3_STANDARD, lazy_fixture("ot3_standard_deck_def")),
    ],
)
async def test_get_deck_definition(
    deck_type: DeckType,
    expected_definition: DeckDefinitionV5,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should be able to load the correct deck definition."""
    subject = DeckDataProvider(
        deck_type=deck_type, labware_data=mock_labware_data_provider
    )
    result = await subject.get_deck_definition()
    assert result == expected_definition


async def test_get_deck_labware_fixtures_ot2_standard(
    decoy: Decoy,
    ot2_standard_deck_def: DeckDefinitionV5,
    ot2_fixed_trash_def: LabwareDefinition,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should be able to get a list of prepopulated labware on the deck."""
    subject = DeckDataProvider(
        deck_type=DeckType.OT2_STANDARD, labware_data=mock_labware_data_provider
    )

    decoy.when(
        await mock_labware_data_provider.get_labware_definition(
            load_name="opentrons_1_trash_1100ml_fixed",
            namespace="opentrons",
            version=1,
        )
    ).then_return(ot2_fixed_trash_def)

    result = await subject.get_deck_fixed_labware(True, ot2_standard_deck_def, None)

    assert result == [
        DeckFixedLabware(
            labware_id="fixedTrash",
            location=DeckSlotLocation(slotName=DeckSlotName.FIXED_TRASH),
            definition=ot2_fixed_trash_def,
        )
    ]


async def test_get_deck_labware_fixtures_ot2_short_trash(
    decoy: Decoy,
    ot2_short_trash_deck_def: DeckDefinitionV5,
    ot2_short_fixed_trash_def: LabwareDefinition,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should be able to get a list of prepopulated labware on the deck."""
    subject = DeckDataProvider(
        deck_type=DeckType.OT2_SHORT_TRASH, labware_data=mock_labware_data_provider
    )

    decoy.when(
        await mock_labware_data_provider.get_labware_definition(
            load_name="opentrons_1_trash_850ml_fixed",
            namespace="opentrons",
            version=1,
        )
    ).then_return(ot2_short_fixed_trash_def)

    result = await subject.get_deck_fixed_labware(True, ot2_short_trash_deck_def, None)

    assert result == [
        DeckFixedLabware(
            labware_id="fixedTrash",
            location=DeckSlotLocation(slotName=DeckSlotName.FIXED_TRASH),
            definition=ot2_short_fixed_trash_def,
        )
    ]


async def test_get_deck_labware_fixtures_ot3_standard(
    decoy: Decoy,
    ot3_standard_deck_def: DeckDefinitionV5,
    ot3_fixed_trash_def: LabwareDefinition,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should be able to get a list of prepopulated labware on the deck."""
    subject = DeckDataProvider(
        deck_type=DeckType.OT3_STANDARD, labware_data=mock_labware_data_provider
    )

    decoy.when(
        await mock_labware_data_provider.get_labware_definition(
            load_name="opentrons_1_trash_3200ml_fixed",
            namespace="opentrons",
            version=1,
        )
    ).then_return(ot3_fixed_trash_def)

    result = await subject.get_deck_fixed_labware(True, ot3_standard_deck_def, None)

    assert result == [
        DeckFixedLabware(
            labware_id="fixedTrash",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A3),
            definition=ot3_fixed_trash_def,
        )
    ]



"""Test magnetic module disengage commands."""

from decoy import Decoy

from opentrons.hardware_control.modules import MagDeck
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.magnetic_module import (
    DisengageParams,
    DisengageResult,
)
from opentrons.protocol_engine.commands.magnetic_module.disengage import (
    DisengageImplementation,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state.module_substates import (
    MagneticModuleId,
    MagneticModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView


async def test_magnetic_module_disengage_implementation(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should validate, find hardware module if not virtualized, and disengage."""
    subject = DisengageImplementation(state_view=state_view, equipment=equipment)

    params = DisengageParams(moduleId="unverified-module-id")

    verified_module_id = MagneticModuleId("module-id")
    magnetic_module_substate = decoy.mock(cls=MagneticModuleSubState)
    magnetic_module_hw = decoy.mock(cls=MagDeck)

    decoy.when(
        state_view.modules.get_magnetic_module_substate("unverified-module-id")
    ).then_return(magnetic_module_substate)

    decoy.when(magnetic_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        magnetic_module_hw
    )

    result = await subject.execute(params=params)

    decoy.verify(await magnetic_module_hw.deactivate(), times=1)
    assert result == SuccessData(public=DisengageResult())



"""Test Flex Stacker unsafe manual retrieve command implementation."""

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
)

from opentrons.hardware_control.modules import FlexStacker
from opentrons.hardware_control.modules.types import PlatformState
from opentrons.protocol_engine.commands import unsafe
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.unsafe.unsafe_stacker_manual_retrieve import (
    UnsafeFlexStackerManualRetrieveImpl,
)
from opentrons.protocol_engine.errors import CannotPerformModuleAction
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    AddressableAreaUsedUpdate,
    BatchLabwareLocationUpdate,
    FlexStackerStateUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    InStackerHopperLocation,
    LabwareLocationSequence,
    LabwareUri,
    LoadedModule,
    ModuleLocation,
    ModuleModel,
    OnAddressableAreaLocationSequenceComponent,
    OnCutoutFixtureLocationSequenceComponent,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
    OnModuleLocationSequenceComponent,
    StackerStoredLabwareGroup,
)
from opentrons.types import DeckSlotName


def _contained_labware(
    count: int, with_adapter: bool = False, with_lid: bool = False
) -> list[StackerStoredLabwareGroup]:
    return [
        StackerStoredLabwareGroup(
            primaryLabwareId=f"primary-id-{i + 1}",
            adapterLabwareId=None if not with_adapter else f"adapter-id-{i + 1}",
            lidLabwareId=None if not with_lid else f"lid-id-{i + 1}",
        )
        for i in range(count)
    ]


@pytest.fixture
def stacker_id() -> FlexStackerId:
    """Get a consistent ID for a stacker."""
    return FlexStackerId("flex-stacker-id")


@pytest.fixture
def stacker_hardware(
    decoy: Decoy, equipment: EquipmentHandler, stacker_id: FlexStackerId
) -> FlexStacker:
    """Get a mocked hardware stacker."""
    hardware = decoy.mock(cls=FlexStacker)
    decoy.when(equipment.get_module_hardware_api(stacker_id)).then_return(hardware)
    return hardware


def _prep_stacker_own_location(
    decoy: Decoy, state_view: StateView, stacker_id: str
) -> None:
    decoy.when(state_view.modules.get_location(stacker_id)).then_return(
        DeckSlotLocation(slotName=DeckSlotName(value="B3")),
    )
    decoy.when(state_view.modules.get(stacker_id)).then_return(
        LoadedModule(
            id=stacker_id,
            location=DeckSlotLocation(slotName=DeckSlotName(value="B3")),
            model=ModuleModel.FLEX_STACKER_MODULE_V1,
            serialNumber="HIIIII",
        )
    )
    decoy.when(
        state_view.modules.ensure_and_convert_module_fixture_location(
            deck_slot=DeckSlotName("B3"), model=ModuleModel.FLEX_STACKER_MODULE_V1
        )
    ).then_return("flexStackerV1B4")


def _stacker_base_loc_seq(stacker_id: str) -> LabwareLocationSequence:
    return [
        OnAddressableAreaLocationSequenceComponent(
            addressableAreaName="flexStackerV1B4"
        ),
        OnModuleLocationSequenceComponent(moduleId=stacker_id),
        OnCutoutFixtureLocationSequenceComponent(
            cutoutId="cutoutB3", possibleCutoutFixtureIds=["flexStackerModuleV1"]
        ),
    ]


@pytest.fixture
def subject(
    state_view: StateView, equipment: EquipmentHandler, model_utils: ModelUtils
) -> UnsafeFlexStackerManualRetrieveImpl:
    """Get a retrieve command to test."""
    return UnsafeFlexStackerManualRetrieveImpl(
        state_view=state_view, equipment=equipment, model_utils=model_utils
    )


async def test_manual_retrieve_raises_when_empty(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerManualRetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    stacker_id: FlexStackerId,
) -> None:
    """It should raise an exception when called on an empty pool."""
    data = unsafe.UnsafeFlexStackerManualRetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=[],
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    with pytest.raises(
        CannotPerformModuleAction,
        match="Cannot retrieve labware from Flex Stacker in .* because it contains no labware",
    ):
        await subject.execute(data)


async def test_manual_retrieve_primary_only(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerManualRetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware."""
    data = unsafe.UnsafeFlexStackerManualRetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(2),
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            InStackerHopperLocation(moduleId=stacker_id),
        )
    ).then_return([InStackerHopperLocation(moduleId=stacker_id)])
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(flex_50uL_tiprack)
    ).then_return("opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    _prep_stacker_own_location(decoy, state_view, stacker_id)

    decoy.when(stacker_hardware.platform_state).then_return(PlatformState.EXTENDED)

    result = await subject.execute(data)

    assert result == SuccessData(
        public=unsafe.UnsafeFlexStackerManualRetrieveResult(
            labwareId="primary-id-1",
            primaryLocationSequence=_stacker_base_loc_seq(stacker_id),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            originalPrimaryLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "primary-id-1": ModuleLocation(moduleId=stacker_id)
                },
                new_offset_ids_by_id={"primary-id-1": None},
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[_contained_labware(2)[1]],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


async def test_manual_retrieve_primary_and_lid(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerManualRetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    tiprack_lid_def: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware with a lid on it."""
    data = unsafe.UnsafeFlexStackerManualRetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=tiprack_lid_def,
        contained_labware_bottom_first=_contained_labware(2, with_lid=True),
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            InStackerHopperLocation(moduleId=stacker_id)
        )
    ).then_return([InStackerHopperLocation(moduleId=stacker_id)])

    _prep_stacker_own_location(decoy, state_view, stacker_id)
    decoy.when(stacker_hardware.platform_state).then_return(PlatformState.EXTENDED)
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(flex_50uL_tiprack)
    ).then_return("opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_lid_def)
    ).then_return("opentrons/opentrons_flex_tiprack_lid/1")

    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    decoy.when(state_view.labware.get_uri_from_definition(tiprack_lid_def)).then_return(
        LabwareUri("opentrons/opentrons_flex_tiprack_lid/1")
    )

    result = await subject.execute(data)

    assert result == SuccessData(
        public=unsafe.UnsafeFlexStackerManualRetrieveResult(
            labwareId="primary-id-1",
            lidId="lid-id-1",
            primaryLocationSequence=_stacker_base_loc_seq(stacker_id),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            lidLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="primary-id-1", lidId="lid-id-1"
                    )
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            lidLabwareURI="opentrons/opentrons_flex_tiprack_lid/1",
            originalPrimaryLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
            originalLidLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="primary-id-1", lidId="lid-id-1"
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "primary-id-1": ModuleLocation(moduleId=stacker_id),
                    "lid-id-1": OnLabwareLocation(labwareId="primary-id-1"),
                },
                new_offset_ids_by_id={"primary-id-1": None, "lid-id-1": None},
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[
                    _contained_labware(2, with_lid=True)[1]
                ],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


async def test_manual_retrieve_primary_and_adapter(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerManualRetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    tiprack_adapter_def: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware on an adapter."""
    data = unsafe.UnsafeFlexStackerManualRetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=tiprack_adapter_def,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(2, with_adapter=True),
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            InStackerHopperLocation(moduleId=stacker_id)
        )
    ).then_return([InStackerHopperLocation(moduleId=stacker_id)])

    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(flex_50uL_tiprack)
    ).then_return("opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_adapter_def)
    ).then_return("opentrons/opentrons_flex_96_tiprack_adapter/1")

    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    decoy.when(
        state_view.labware.get_uri_from_definition(tiprack_adapter_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_tiprack_adapter/1"))

    _prep_stacker_own_location(decoy, state_view, stacker_id)
    decoy.when(stacker_hardware.platform_state).then_return(PlatformState.EXTENDED)

    result = await subject.execute(data)

    assert result == SuccessData(
        public=unsafe.UnsafeFlexStackerManualRetrieveResult(
            labwareId="primary-id-1",
            adapterId="adapter-id-1",
            primaryLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    )
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            adapterLocationSequence=_stacker_base_loc_seq(stacker_id),
            adapterLabwareURI="opentrons/opentrons_flex_96_tiprack_adapter/1",
            originalPrimaryLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="adapter-id-1", lidId=None
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
            originalAdapterLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "adapter-id-1": ModuleLocation(moduleId=stacker_id),
                    "primary-id-1": OnLabwareLocation(labwareId="adapter-id-1"),
                },
                new_offset_ids_by_id={"primary-id-1": None, "adapter-id-1": None},
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[
                    _contained_labware(2, with_adapter=True)[1]
                ],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


async def test_manual_retrieve_primary_adapter_and_lid(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerManualRetrieveImpl,
    flex_50uL_tiprack: LabwareDefinition,
    tiprack_adapter_def: LabwareDefinition,
    tiprack_lid_def: LabwareDefinition,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should be able to retrieve a labware on an adapter."""
    data = unsafe.UnsafeFlexStackerManualRetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=tiprack_adapter_def,
        pool_lid_definition=tiprack_lid_def,
        contained_labware_bottom_first=_contained_labware(
            2, with_adapter=True, with_lid=True
        ),
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            InStackerHopperLocation(moduleId=stacker_id)
        )
    ).then_return([InStackerHopperLocation(moduleId=stacker_id)])
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            OnLabwareLocation(labwareId="adapter-id-1"),
        )
    ).then_return(
        [OnLabwareLocationSequenceComponent(labwareId="adapter-id-1", lidId=None)]
        + _stacker_base_loc_seq(stacker_id)
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            OnLabwareLocation(labwareId="primary-id-1"),
        )
    ).then_return(
        [
            OnLabwareLocationSequenceComponent(
                labwareId="primary-id-1", lidId="lid-id-1"
            ),
            OnLabwareLocationSequenceComponent(labwareId="adapter-id-1", lidId=None),
        ]
        + _stacker_base_loc_seq(stacker_id)
    )
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(flex_50uL_tiprack)
    ).then_return("opentrons/opentrons_flex_96_filtertiprack_50ul/1")
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_adapter_def)
    ).then_return("opentrons/opentrons_flex_96_tiprack_adapter/1")
    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(tiprack_lid_def)
    ).then_return("opentrons/opentrons_flex_tiprack_lid/1")

    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    decoy.when(
        state_view.labware.get_uri_from_definition(tiprack_adapter_def)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_tiprack_adapter/1"))
    decoy.when(state_view.labware.get_uri_from_definition(tiprack_lid_def)).then_return(
        LabwareUri("opentrons/opentrons_flex_tiprack_lid/1")
    )

    _prep_stacker_own_location(decoy, state_view, stacker_id)
    decoy.when(stacker_hardware.platform_state).then_return(PlatformState.EXTENDED)

    result = await subject.execute(data)

    assert result == SuccessData(
        public=unsafe.UnsafeFlexStackerManualRetrieveResult(
            labwareId="primary-id-1",
            adapterId="adapter-id-1",
            lidId="lid-id-1",
            primaryLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    )
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
            adapterLocationSequence=_stacker_base_loc_seq(stacker_id),
            adapterLabwareURI="opentrons/opentrons_flex_96_tiprack_adapter/1",
            lidLocationSequence=(
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="primary-id-1", lidId="lid-id-1"
                    ),
                    OnLabwareLocationSequenceComponent(
                        labwareId="adapter-id-1", lidId=None
                    ),
                ]
                + _stacker_base_loc_seq(stacker_id)
            ),
            lidLabwareURI="opentrons/opentrons_flex_tiprack_lid/1",
            originalPrimaryLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="adapter-id-1", lidId=None
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
            originalAdapterLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id)
            ],
            originalLidLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="primary-id-1", lidId="lid-id-1"
                ),
                OnLabwareLocationSequenceComponent(
                    labwareId="adapter-id-1", lidId=None
                ),
                InStackerHopperLocation(moduleId=stacker_id),
            ],
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "adapter-id-1": ModuleLocation(moduleId=stacker_id),
                    "primary-id-1": OnLabwareLocation(labwareId="adapter-id-1"),
                    "lid-id-1": OnLabwareLocation(labwareId="primary-id-1"),
                },
                new_offset_ids_by_id={
                    "primary-id-1": None,
                    "adapter-id-1": None,
                    "lid-id-1": None,
                },
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=[
                    _contained_labware(2, with_adapter=True, with_lid=True)[1]
                ],
            ),
            addressable_area_used=AddressableAreaUsedUpdate(
                addressable_area_name="flexStackerV1B4"
            ),
        ),
    )


async def test_manual_retrieve_fails_due_to_platform_state(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    subject: UnsafeFlexStackerManualRetrieveImpl,
    model_utils: ModelUtils,
    stacker_id: FlexStackerId,
    flex_50uL_tiprack: LabwareDefinition,
    stacker_hardware: FlexStacker,
) -> None:
    """It should raise a CannotPerformModuleAction error."""
    data = unsafe.UnsafeFlexStackerManualRetrieveParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(1),
        max_pool_count=999,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id),
        )
    ).then_return(_stacker_base_loc_seq(stacker_id))
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            InStackerHopperLocation(moduleId=stacker_id)
        )
    ).then_return([InStackerHopperLocation(moduleId=stacker_id)])

    decoy.when(stacker_hardware.platform_state).then_return(PlatformState.UNKNOWN)
    with pytest.raises(
        CannotPerformModuleAction,
        match="Cannot manually retrieve a labware from Flex Stacker in .* if the carriage is not in gripper position.",
    ):
        await subject.execute(data)



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import pytest
from mock import MagicMock, PropertyMock, call
from serial import Serial  # type: ignore[import-untyped]

from opentrons.drivers.asyncio.communication import AsyncSerial


@pytest.fixture
def mock_timeout_prop() -> PropertyMock:
    return PropertyMock()


@pytest.fixture
def mock_write_timeout_prop() -> PropertyMock:
    return PropertyMock()


@pytest.fixture
def mock_serial(
    mock_timeout_prop: PropertyMock, mock_write_timeout_prop: PropertyMock
) -> MagicMock:
    """Mock Serial"""
    m = MagicMock(spec=Serial)
    type(m).timeout = mock_timeout_prop
    type(m).write_timeout = mock_write_timeout_prop
    return m


@pytest.fixture
async def subject(mock_serial: MagicMock) -> AsyncSerial:
    """The test subject."""
    return AsyncSerial(
        serial=mock_serial,
        executor=ThreadPoolExecutor(),
        loop=asyncio.get_running_loop(),
        reset_buffer_before_write=False,
    )


@pytest.mark.parametrize(
    argnames=["default", "override"],
    argvalues=[
        [None, 5],
        [5, 6],
    ],
)
async def test_write_timeout_override(
    subject: AsyncSerial,
    mock_write_timeout_prop: PropertyMock,
    default: Optional[int],
    override: Optional[int],
) -> None:
    """It should override the timeout and return to default after the call."""
    mock_write_timeout_prop.return_value = default
    async with subject.timeout_override("write_timeout", override):
        await subject.write(b"")

    # Three calls: read, override, reset default.
    assert mock_write_timeout_prop.call_args_list == [
        call(),
        call(override),
        call(default),
    ]


@pytest.mark.parametrize(
    argnames=["default", "override"],
    argvalues=[
        [None, 5],
        [5, 6],
    ],
)
async def test_timeout_override(
    subject: AsyncSerial,
    mock_timeout_prop: PropertyMock,
    default: Optional[int],
    override: Optional[int],
) -> None:
    """It should override the timeout and return to default after the call."""
    mock_timeout_prop.return_value = default
    async with subject.timeout_override("timeout", override):
        await subject.read_until(b"")

    # Three calls: read, override, reset default.
    assert mock_timeout_prop.call_args_list == [call(), call(override), call(default)]


@pytest.mark.parametrize(
    argnames=["default", "override"],
    argvalues=[
        [5, 5],
        [5, None],
        [None, None],
    ],
)
async def test_write_timeout_dont_override(
    subject: AsyncSerial,
    mock_write_timeout_prop: PropertyMock,
    default: Optional[int],
    override: Optional[int],
) -> None:
    """It should not override the timeout if not necessary."""
    mock_write_timeout_prop.return_value = default
    async with subject.timeout_override("write_timeout", override):
        await subject.write(b"")

    mock_write_timeout_prop.assert_called_once()


@pytest.mark.parametrize(
    argnames=["default", "override"],
    argvalues=[
        [5, 5],
        [5, None],
        [None, None],
    ],
)
async def test_read_timeout_dont_override(
    subject: AsyncSerial,
    mock_timeout_prop: PropertyMock,
    default: Optional[int],
    override: Optional[int],
) -> None:
    """It should not override the timeout if not necessary."""
    mock_timeout_prop.return_value = default
    async with subject.timeout_override("timeout", override):
        await subject.read_until(b"")

    mock_timeout_prop.assert_called_once()


def test_reset_input_buffer(mock_serial: MagicMock, subject: AsyncSerial) -> None:
    """It should call the underlying serial port's Reset function"""
    subject.reset_input_buffer()
    mock_serial.reset_input_buffer.assert_called_once()


@pytest.mark.parametrize(
    "flush_side_effect",
    [
        None,
        OSError("device disconnected!"),
    ],
)
async def test_no_exception_on_flush(
    mock_serial: MagicMock, subject: AsyncSerial, flush_side_effect: Optional[Exception]
) -> None:
    """It should ignore exceptions from flush.

    This can happen after a device has been issued a `dfu` command
    and disconnects before the flush command executes.
    """
    mock_serial.flush.side_effect = flush_side_effect

    try:
        await subject.write("dfu".encode())
    except Exception as e:
        pytest.fail(f"_sync_write leaked an exception: {e!r}")

    mock_serial.flush.assert_called_once()
    mock_serial.reset_mock()



from . import types as FlexStackerTypes
from . import utils as FlexStackerUtils
from .abstract import AbstractFlexStackerDriver
from .driver import FlexStackerDriver
from .simulator import SimulatingDriver

__all__ = [
    "AbstractFlexStackerDriver",
    "FlexStackerDriver",
    "SimulatingDriver",
    "FlexStackerTypes",
    "FlexStackerUtils",
]



"""Test touch tip commands."""

import pytest
from decoy import Decoy

from opentrons.hardware_control.types import CriticalPoint
from opentrons.motion_planning import Waypoint
from opentrons.protocol_engine import DeckPoint, WellLocation, WellOffset, errors
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.touch_tip import (
    TouchTipImplementation,
    TouchTipParams,
    TouchTipResult,
)
from opentrons.protocol_engine.execution import GantryMover, MovementHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId
from opentrons.types import Point


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def mock_model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mock ModelUtils."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def mock_movement_handler(decoy: Decoy) -> MovementHandler:
    """Get a mock MovementHandler."""
    return decoy.mock(cls=MovementHandler)


@pytest.fixture
def mock_gantry_mover(decoy: Decoy) -> GantryMover:
    """Get a mock GantryMover."""
    return decoy.mock(cls=GantryMover)


@pytest.fixture
def subject(
    mock_state_view: StateView,
    mock_movement_handler: MovementHandler,
    mock_gantry_mover: GantryMover,
    mock_model_utils: ModelUtils,
) -> TouchTipImplementation:
    """Get the test subject."""
    return TouchTipImplementation(
        state_view=mock_state_view,
        movement=mock_movement_handler,
        gantry_mover=mock_gantry_mover,
        model_utils=mock_model_utils,
    )


async def test_touch_tip_implementation(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_movement_handler: MovementHandler,
    mock_gantry_mover: GantryMover,
    subject: TouchTipImplementation,
) -> None:
    """A TouchTip command should have an execution implementation."""
    params = TouchTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=WellLocation(offset=WellOffset(x=1, y=2, z=3)),
        radius=0.456,
        speed=42.0,
    )

    decoy.when(
        await mock_movement_handler.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=1, y=2, z=3)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        mock_state_view.pipettes.get_movement_speed(
            pipette_id="abc", requested_speed=42.0
        )
    ).then_return(9001)

    decoy.when(
        mock_state_view.motion.get_touch_tip_waypoints(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            radius=0.456,
            mm_from_edge=0,
            center_point=Point(x=1, y=2, z=3),
        )
    ).then_return(
        [
            Waypoint(
                position=Point(x=11, y=22, z=33),
                critical_point=CriticalPoint.XY_CENTER,
            ),
            Waypoint(
                position=Point(x=44, y=55, z=66),
                critical_point=CriticalPoint.XY_CENTER,
            ),
        ]
    )

    decoy.when(
        await mock_gantry_mover.move_to(
            pipette_id="abc",
            waypoints=[
                Waypoint(
                    position=Point(x=11, y=22, z=33),
                    critical_point=CriticalPoint.XY_CENTER,
                ),
                Waypoint(
                    position=Point(x=44, y=55, z=66),
                    critical_point=CriticalPoint.XY_CENTER,
                ),
            ],
            speed=9001,
        )
    ).then_return(Point(x=4, y=5, z=6))

    result = await subject.execute(params)

    assert result == SuccessData(
        public=TouchTipResult(position=DeckPoint(x=4, y=5, z=6)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(labware_id="123", well_name="A3"),
                new_deck_point=DeckPoint(x=4, y=5, z=6),
            )
        ),
    )


async def test_touch_tip_implementation_with_mm_from_edge(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_movement_handler: MovementHandler,
    mock_gantry_mover: GantryMover,
    subject: TouchTipImplementation,
) -> None:
    """A TouchTip command should use mmFromEdge if provided."""
    params = TouchTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=WellLocation(offset=WellOffset(x=1, y=2, z=3)),
        mmFromEdge=0.789,
        speed=42.0,
    )

    decoy.when(
        await mock_movement_handler.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=1, y=2, z=3)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        mock_state_view.pipettes.get_movement_speed(
            pipette_id="abc", requested_speed=42.0
        )
    ).then_return(9001)

    decoy.when(
        mock_state_view.motion.get_touch_tip_waypoints(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            radius=1.0,
            mm_from_edge=0.789,
            center_point=Point(x=1, y=2, z=3),
        )
    ).then_return(
        [
            Waypoint(
                position=Point(x=11, y=22, z=33),
                critical_point=CriticalPoint.XY_CENTER,
            ),
            Waypoint(
                position=Point(x=44, y=55, z=66),
                critical_point=CriticalPoint.XY_CENTER,
            ),
        ]
    )

    decoy.when(
        await mock_gantry_mover.move_to(
            pipette_id="abc",
            waypoints=[
                Waypoint(
                    position=Point(x=11, y=22, z=33),
                    critical_point=CriticalPoint.XY_CENTER,
                ),
                Waypoint(
                    position=Point(x=44, y=55, z=66),
                    critical_point=CriticalPoint.XY_CENTER,
                ),
            ],
            speed=9001,
        )
    ).then_return(Point(x=4, y=5, z=6))

    result = await subject.execute(params)

    assert result == SuccessData(
        public=TouchTipResult(position=DeckPoint(x=4, y=5, z=6)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(labware_id="123", well_name="A3"),
                new_deck_point=DeckPoint(x=4, y=5, z=6),
            )
        ),
    )


async def test_touch_tip_disabled(
    decoy: Decoy, mock_state_view: StateView, subject: TouchTipImplementation
) -> None:
    """It should disallow touch tip on labware with the touchTipDisabled quirk."""
    params = TouchTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=WellLocation(),
    )

    decoy.when(
        mock_state_view.labware.get_has_quirk("123", "touchTipDisabled")
    ).then_return(True)

    with pytest.raises(errors.TouchTipDisabledError):
        await subject.execute(params)


async def test_touch_tip_no_tip_racks(
    decoy: Decoy, mock_state_view: StateView, subject: TouchTipImplementation
) -> None:
    """It should disallow touch tip on tip racks."""
    params = TouchTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=WellLocation(),
    )

    decoy.when(mock_state_view.labware.is_tiprack("123")).then_return(True)

    with pytest.raises(errors.LabwareIsTipRackError):
        await subject.execute(params)


async def test_touch_tip_incompatible_arguments(
    decoy: Decoy, mock_state_view: StateView, subject: TouchTipImplementation
) -> None:
    """It should disallow touch tip if radius and mmFromEdge is provided."""
    params = TouchTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=WellLocation(),
        radius=1.23,
        mmFromEdge=4.56,
    )

    with pytest.raises(errors.TouchTipIncompatibleArgumentsError):
        await subject.execute(params)



"""Test the ``moveLabware`` command."""

import inspect
from datetime import datetime
from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    EnumeratedError,
    FailedGripperPickupError,
    LabwareDroppedError,
    StallOrCollisionDetectedError,
)
from opentrons_shared_data.gripper.constants import GRIPPER_PADDLE_WIDTH
from opentrons_shared_data.labware.labware_definition import (
    Dimensions,
    LabwareDefinition2,
    Parameters2,
)

from opentrons.protocol_engine import Config, errors
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.move_labware import (
    GripperMovementError,
    MoveLabwareImplementation,
    MoveLabwareParams,
    MoveLabwareResult,
)
from opentrons.protocol_engine.execution import (
    EquipmentHandler,
    LabwareMovementHandler,
    RunControlHandler,
)
from opentrons.protocol_engine.resources import labware_validation
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import (
    WASTE_CHUTE_LOCATION,
    AddressableAreaLocation,
    CurrentWell,
    DeckSlotLocation,
    DeckType,
    LabwareLocationSequence,
    LabwareMovementStrategy,
    LabwareOffsetVector,
    LoadedLabware,
    ModuleLocation,
    NotOnDeckLocationSequenceComponent,
    OnAddressableAreaLocationSequenceComponent,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
)
from opentrons.types import DeckSlotName, Point


@pytest.fixture(autouse=True)
def patch_mock_labware_validation(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out labware_validation.py functions."""
    for name, func in inspect.getmembers(labware_validation, inspect.isfunction):
        monkeypatch.setattr(labware_validation, name, decoy.mock(func=func))


@pytest.fixture
def subject(
    equipment: EquipmentHandler,
    labware_movement: LabwareMovementHandler,
    state_view: StateView,
    run_control: RunControlHandler,
    model_utils: ModelUtils,
) -> MoveLabwareImplementation:
    """Return a test subject configured to use mocked-out dependencies."""
    return MoveLabwareImplementation(
        state_view=state_view,
        equipment=equipment,
        labware_movement=labware_movement,
        run_control=run_control,
        model_utils=model_utils,
    )


@pytest.mark.parametrize(
    argnames=["strategy", "times_pause_called"],
    argvalues=[
        [LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE, 1],
        [LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE, 0],
    ],
)
async def test_manual_move_labware_implementation(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    equipment: EquipmentHandler,
    state_view: StateView,
    run_control: RunControlHandler,
    strategy: LabwareMovementStrategy,
    times_pause_called: int,
) -> None:
    """It should execute a pause and return the new offset."""
    new_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_4)
    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=new_location,
        strategy=strategy,
    )

    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
        )
    ).then_return(DeckSlotLocation(slotName=DeckSlotName.SLOT_5))
    decoy.when(
        equipment.find_applicable_labware_offset_id(
            labware_definition_uri="opentrons-test/load-name/1",
            labware_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        )
    ).then_return("wowzers-a-new-offset-id")
    decoy.when(
        state_view.geometry.get_location_sequence("my-cool-labware-id")
    ).then_return([OnAddressableAreaLocationSequenceComponent(addressableAreaName="5")])
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(new_location)
    ).then_return([OnAddressableAreaLocationSequenceComponent(addressableAreaName="4")])

    result = await subject.execute(data)
    decoy.verify(await run_control.wait_for_resume(), times=times_pause_called)
    decoy.verify(
        state_view.labware.raise_if_labware_has_non_lid_labware_on_top(
            "my-cool-labware-id"
        )
    )
    assert result == SuccessData(
        public=MoveLabwareResult(
            offsetId="wowzers-a-new-offset-id",
            originLocationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="5")
            ],
            immediateDestinationLocationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="4")
            ],
            eventualDestinationLocationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="4")
            ],
        ),
        state_update=update_types.StateUpdate(
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="my-cool-labware-id",
                offset_id="wowzers-a-new-offset-id",
                new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
            ),
            addressable_area_used=update_types.AddressableAreaUsedUpdate(
                addressable_area_name=new_location.slotName.id
            ),
        ),
    )


async def test_move_labware_implementation_on_labware(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    equipment: EquipmentHandler,
    state_view: StateView,
    run_control: RunControlHandler,
) -> None:
    """It should execute a pause and return the new offset."""
    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=OnLabwareLocation(labwareId="new-labware-id"),
        strategy=LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE,
    )

    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.labware.get_definition(labware_id="my-cool-labware-id")
    ).then_return(
        LabwareDefinition2.model_construct(namespace="spacename")  # type: ignore[call-arg]
    )
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=OnLabwareLocation(labwareId="new-labware-id"),
        )
    ).then_return(OnLabwareLocation(labwareId="my-even-cooler-labware-id"))
    decoy.when(
        equipment.find_applicable_labware_offset_id(
            labware_definition_uri="opentrons-test/load-name/1",
            labware_location=OnLabwareLocation(labwareId="my-even-cooler-labware-id"),
        )
    ).then_return("wowzers-a-new-offset-id")
    decoy.when(
        state_view.geometry.get_location_sequence("my-cool-labware-id")
    ).then_return([OnAddressableAreaLocationSequenceComponent(addressableAreaName="1")])
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            OnLabwareLocation(labwareId="new-labware-id")
        )
    ).then_return(
        [
            OnLabwareLocationSequenceComponent(labwareId="new-labware-id", lidId=None),
            OnAddressableAreaLocationSequenceComponent(addressableAreaName="2"),
        ]
    )

    result = await subject.execute(data)
    decoy.verify(
        state_view.labware.raise_if_labware_has_non_lid_labware_on_top(
            "my-cool-labware-id"
        ),
        state_view.labware.raise_if_labware_has_labware_on_top(
            "my-even-cooler-labware-id"
        ),
        state_view.labware.raise_if_labware_cannot_be_stacked(
            LabwareDefinition2.model_construct(namespace="spacename"),  # type: ignore[call-arg]
            "my-even-cooler-labware-id",
        ),
    )
    assert result == SuccessData(
        public=MoveLabwareResult(
            offsetId="wowzers-a-new-offset-id",
            originLocationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="1")
            ],
            immediateDestinationLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="new-labware-id", lidId=None
                ),
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="2"),
            ],
            eventualDestinationLocationSequence=[
                OnLabwareLocationSequenceComponent(
                    labwareId="new-labware-id", lidId=None
                ),
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="2"),
            ],
        ),
        state_update=update_types.StateUpdate(
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="my-cool-labware-id",
                offset_id="wowzers-a-new-offset-id",
                new_location=OnLabwareLocation(labwareId="my-even-cooler-labware-id"),
            )
        ),
    )


async def test_gripper_move_labware_implementation(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    equipment: EquipmentHandler,
    labware_movement: LabwareMovementHandler,
    state_view: StateView,
) -> None:
    """It should delegate to the equipment handler and return the new offset."""
    from_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    new_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_5)
    pick_up_offset = LabwareOffsetVector(x=1, y=2, z=3)

    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=new_location,
        strategy=LabwareMovementStrategy.USING_GRIPPER,
        pickUpOffset=pick_up_offset,
        dropOffset=None,
    )

    decoy.when(
        state_view.labware.get_definition(labware_id="my-cool-labware-id")
    ).then_return(sentinel.labware_definition)
    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=from_location,
            offsetId=None,
        )
    )
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(location=new_location)
    ).then_return(sentinel.new_location_validated_unoccupied)
    decoy.when(
        equipment.find_applicable_labware_offset_id(
            labware_definition_uri="opentrons-test/load-name/1",
            labware_location=sentinel.new_location_validated_unoccupied,
        )
    ).then_return("wowzers-a-new-offset-id")

    decoy.when(
        state_view.geometry.ensure_valid_gripper_location(from_location)
    ).then_return(sentinel.from_location_validated_for_gripper)
    decoy.when(
        state_view.geometry.ensure_valid_gripper_location(
            sentinel.new_location_validated_unoccupied
        )
    ).then_return(sentinel.new_location_validated_for_gripper)
    decoy.when(
        labware_validation.validate_gripper_compatible(sentinel.labware_definition)
    ).then_return(True)
    decoy.when(
        state_view.geometry.get_location_sequence("my-cool-labware-id")
    ).then_return(
        [
            OnAddressableAreaLocationSequenceComponent(
                addressableAreaName="1",
            )
        ]
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            sentinel.new_location_validated_for_gripper
        )
    ).then_return([OnAddressableAreaLocationSequenceComponent(addressableAreaName="5")])

    result = await subject.execute(data)
    decoy.verify(
        state_view.labware.raise_if_labware_has_non_lid_labware_on_top(
            "my-cool-labware-id"
        ),
        await labware_movement.move_labware_with_gripper(
            labware_id="my-cool-labware-id",
            current_location=sentinel.from_location_validated_for_gripper,
            new_location=sentinel.new_location_validated_for_gripper,
            user_pick_up_offset=Point(
                pick_up_offset.x, pick_up_offset.y, pick_up_offset.z
            ),
            user_drop_offset=Point(),
            post_drop_slide_offset=None,
        ),
    )
    assert result == SuccessData(
        public=MoveLabwareResult(
            offsetId="wowzers-a-new-offset-id",
            originLocationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="1")
            ],
            immediateDestinationLocationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="5")
            ],
            eventualDestinationLocationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="5")
            ],
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="my-cool-labware-id",
                new_location=sentinel.new_location_validated_unoccupied,
                offset_id="wowzers-a-new-offset-id",
            ),
            addressable_area_used=update_types.AddressableAreaUsedUpdate(
                addressable_area_name=new_location.slotName.id
            ),
        ),
    )


@pytest.mark.parametrize(
    "underlying_exception",
    [
        FailedGripperPickupError(),
        LabwareDroppedError(),
        StallOrCollisionDetectedError(),
    ],
)
async def test_gripper_error(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
    model_utils: ModelUtils,
    labware_movement: LabwareMovementHandler,
    underlying_exception: EnumeratedError,
) -> None:
    """Test the handling of errors during a gripper movement."""
    labware_id = "labware-id"
    labware_namespace = "labware-namespace"
    labware_load_name = "load-name"
    labware_definition_uri = "opentrons-test/load-name/1"
    labware_def = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        namespace=labware_namespace,
    )
    origin_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_A1)
    new_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_A2)
    error_id = "error-id"
    error_created_at = datetime.now()

    # Common MoveLabwareImplementation boilerplate:
    decoy.when(state_view.labware.get_definition(labware_id=labware_id)).then_return(
        LabwareDefinition2.model_construct(namespace=labware_namespace)  # type: ignore[call-arg]
    )
    decoy.when(state_view.labware.get(labware_id=labware_id)).then_return(
        LoadedLabware(
            id=labware_id,
            loadName=labware_load_name,
            definitionUri=labware_definition_uri,
            location=origin_location,
            offsetId=None,
        )
    )
    decoy.when(
        state_view.geometry.ensure_valid_gripper_location(origin_location)
    ).then_return(origin_location)
    decoy.when(
        state_view.geometry.ensure_valid_gripper_location(new_location)
    ).then_return(new_location)
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=new_location,
        )
    ).then_return(new_location)
    decoy.when(labware_validation.validate_gripper_compatible(labware_def)).then_return(
        True
    )
    params = MoveLabwareParams(
        labwareId=labware_id,
        newLocation=new_location,
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )

    # Actual setup for this test:
    decoy.when(
        await labware_movement.move_labware_with_gripper(  # type: ignore[func-returns-value]
            labware_id=labware_id,
            current_location=origin_location,
            new_location=new_location,
            user_pick_up_offset=Point(),
            user_drop_offset=Point(),
            post_drop_slide_offset=None,
        )
    ).then_raise(underlying_exception)
    decoy.when(model_utils.get_timestamp()).then_return(error_created_at)
    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(state_view.geometry.get_location_sequence("labware-id")).then_return(
        [OnAddressableAreaLocationSequenceComponent(addressableAreaName="A1")]
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(new_location)
    ).then_return(
        [OnAddressableAreaLocationSequenceComponent(addressableAreaName="A2")]
    )

    result = await subject.execute(params)

    assert result == DefinedErrorData(
        public=GripperMovementError.model_construct(
            id=error_id,
            createdAt=error_created_at,
            errorCode=underlying_exception.code.value.code,
            detail=underlying_exception.code.value.detail,
            errorInfo={
                "originLocationSequence": [
                    OnAddressableAreaLocationSequenceComponent(addressableAreaName="A1")
                ],
                "immediateDestinationLocationSequence": [
                    OnAddressableAreaLocationSequenceComponent(addressableAreaName="A2")
                ],
                "eventualDestinationLocationSequence": [
                    OnAddressableAreaLocationSequenceComponent(addressableAreaName="A2")
                ],
            },
            wrappedErrors=[matchers.Anything()],
        ),
        state_update=update_types.StateUpdate(
            labware_location=update_types.NO_CHANGE,
            pipette_location=update_types.CLEAR,
            addressable_area_used=update_types.AddressableAreaUsedUpdate(
                addressable_area_name=new_location.slotName.id
            ),
        ),
    )


@pytest.mark.parametrize(
    ("current_labware_id", "moved_labware_id", "expect_cleared_location"),
    [
        ("lw1", "lw2", False),
        ("lw1", "lw1", True),
    ],
)
async def test_clears_location_if_current_labware_moved_from_under_pipette(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
    current_labware_id: str,
    moved_labware_id: str,
    expect_cleared_location: bool,
) -> None:
    """If it moves the labware that the pipette is currently over, it should clear the location."""
    from_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_A1)
    to_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_A2)

    decoy.when(state_view.labware.get(labware_id=moved_labware_id)).then_return(
        LoadedLabware(
            id=moved_labware_id,
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=from_location,
            offsetId=None,
        )
    )

    decoy.when(state_view.pipettes.get_current_location()).then_return(
        CurrentWell(
            pipette_id="pipette-id", labware_id=current_labware_id, well_name="A1"
        )
    )
    decoy.when(state_view.geometry.get_location_sequence(moved_labware_id)).then_return(
        []
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(to_location)
    ).then_return([])

    result = await subject.execute(
        params=MoveLabwareParams(
            labwareId=moved_labware_id,
            newLocation=to_location,
            strategy=LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE,
        )
    )
    assert (
        result.state_update.pipette_location == update_types.CLEAR
        if expect_cleared_location
        else update_types.NO_CHANGE
    )


async def test_gripper_move_to_waste_chute_implementation(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    equipment: EquipmentHandler,
    labware_movement: LabwareMovementHandler,
    state_view: StateView,
) -> None:
    """It should drop the labware with a delay added."""
    from_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    new_location = AddressableAreaLocation(addressableAreaName="gripperWasteChute")
    labware_width = 50
    expected_slide_offset = Point(
        x=labware_width / 2 + GRIPPER_PADDLE_WIDTH / 2 + 8, y=0, z=0
    )
    from_loc_sequence: LabwareLocationSequence = [
        OnAddressableAreaLocationSequenceComponent(addressableAreaName="1")
    ]
    immediate_dest_loc_sequence: LabwareLocationSequence = [
        NotOnDeckLocationSequenceComponent(logicalLocationName=WASTE_CHUTE_LOCATION)
    ]
    eventual_dest_loc_sequence = immediate_dest_loc_sequence

    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=new_location,
        strategy=LabwareMovementStrategy.USING_GRIPPER,
        pickUpOffset=LabwareOffsetVector(x=1, y=2, z=3),
        dropOffset=None,
    )
    labware_def = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        namespace="my-cool-namespace",
        dimensions=Dimensions(
            yDimension=labware_width, zDimension=labware_width, xDimension=labware_width
        ),
    )
    decoy.when(
        state_view.labware.get_definition(labware_id="my-cool-labware-id")
    ).then_return(labware_def)
    decoy.when(
        state_view.geometry.get_location_sequence("my-cool-labware-id")
    ).then_return(from_loc_sequence)
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(new_location)
    ).then_return(immediate_dest_loc_sequence)
    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=from_location,
            offsetId=None,
        )
    )
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=new_location,
        )
    ).then_return(new_location)
    decoy.when(
        equipment.find_applicable_labware_offset_id(
            labware_definition_uri="opentrons-test/load-name/1",
            labware_location=new_location,
        )
    ).then_return("wowzers-a-new-offset-id")

    decoy.when(
        state_view.geometry.ensure_valid_gripper_location(from_location)
    ).then_return(from_location)
    decoy.when(
        state_view.geometry.ensure_valid_gripper_location(new_location)
    ).then_return(new_location)
    decoy.when(labware_validation.validate_gripper_compatible(labware_def)).then_return(
        True
    )

    result = await subject.execute(data)
    decoy.verify(
        state_view.labware.raise_if_labware_has_non_lid_labware_on_top(
            "my-cool-labware-id"
        ),
        await labware_movement.move_labware_with_gripper(
            labware_id="my-cool-labware-id",
            current_location=from_location,
            new_location=new_location,
            user_pick_up_offset=Point(1, 2, 3),
            user_drop_offset=Point(),
            post_drop_slide_offset=expected_slide_offset,
        ),
    )
    assert result == SuccessData(
        public=MoveLabwareResult(
            offsetId="wowzers-a-new-offset-id",
            originLocationSequence=from_loc_sequence,
            immediateDestinationLocationSequence=immediate_dest_loc_sequence,
            eventualDestinationLocationSequence=eventual_dest_loc_sequence,
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
            labware_location=update_types.LabwareLocationUpdate(
                labware_id="my-cool-labware-id",
                new_location=new_location,
                offset_id="wowzers-a-new-offset-id",
            ),
            addressable_area_used=update_types.AddressableAreaUsedUpdate(
                addressable_area_name=new_location.addressableAreaName
            ),
        ),
    )


async def test_move_labware_raises_for_labware_or_module_not_found(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """It should raise an error when specified labware/ module is not found."""
    move_non_existent_labware_params = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )
    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_raise(
        errors.LabwareNotLoadedError("Woops!")
    )
    with pytest.raises(errors.LabwareNotLoadedError):
        await subject.execute(move_non_existent_labware_params)

    move_labware_from_questionable_module_params = MoveLabwareParams(
        labwareId="real-labware-id",
        newLocation=ModuleLocation(moduleId="imaginary-module-id-1"),
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )
    decoy.when(state_view.labware.get(labware_id="real-labware-id")).then_return(
        LoadedLabware(
            id="real-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=ModuleLocation(moduleId="imaginary-module-id-3"),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=ModuleLocation(moduleId="imaginary-module-id-1"),
        )
    ).then_return(ModuleLocation(moduleId="imaginary-module-id-2"))

    decoy.when(
        equipment.find_applicable_labware_offset_id(
            labware_definition_uri="opentrons-test/load-name/1",
            labware_location=ModuleLocation(moduleId="imaginary-module-id-2"),
        )
    ).then_raise(errors.ModuleNotLoadedError(module_id="woops-i-dont-exist"))

    with pytest.raises(errors.ModuleNotLoadedError):
        await subject.execute(move_labware_from_questionable_module_params)


async def test_move_labware_raises_if_movement_obstructed(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    equipment: EquipmentHandler,
    labware_movement: LabwareMovementHandler,
    state_view: StateView,
) -> None:
    """It should execute a pause and return the new offset."""
    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        strategy=LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE,
    )
    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        )
    ).then_return(DeckSlotLocation(slotName=DeckSlotName.SLOT_6))
    decoy.when(
        equipment.find_applicable_labware_offset_id(
            labware_definition_uri="opentrons-test/load-name/1",
            labware_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_6),
        )
    ).then_return("wowzers-a-new-offset-id")

    decoy.when(
        await labware_movement.ensure_movement_not_obstructed_by_module(  # type: ignore[func-returns-value]
            labware_id="my-cool-labware-id",
            new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_6),
        )
    ).then_raise(errors.LabwareMovementNotAllowedError("Oh boy"))

    with pytest.raises(errors.LabwareMovementNotAllowedError):
        await subject.execute(data)


async def test_move_labware_raises_when_location_occupied(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
) -> None:
    """It should raise an error when trying to move labware to non-empty location."""
    move_labware_params = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )
    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        )
    ).then_raise(errors.LocationIsOccupiedError("Woops!"))

    with pytest.raises(errors.LocationIsOccupiedError):
        await subject.execute(move_labware_params)


async def test_move_labware_raises_when_moving_adapter_with_gripper(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
) -> None:
    """It should raise an error when trying to move an adapter with a gripper."""
    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )

    definition = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        parameters=Parameters2.model_construct(loadName="My cool adapter"),  # type: ignore[call-arg]
    )

    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.labware.get_definition(labware_id="my-cool-labware-id")
    ).then_return(definition)
    decoy.when(labware_validation.validate_gripper_compatible(definition)).then_return(
        True
    )
    decoy.when(
        labware_validation.validate_definition_is_adapter(definition)
    ).then_return(True)

    with pytest.raises(
        errors.LabwareMovementNotAllowedError, match="move adapter 'My cool adapter'"
    ):
        await subject.execute(data)


async def test_move_labware_raises_when_moving_labware_with_gripper_incompatible_quirk(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
) -> None:
    """It should raise an error when trying to move an adapter with a gripper."""
    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )

    definition = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        parameters=Parameters2.model_construct(loadName="My cool labware"),  # type: ignore[call-arg]
    )

    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.labware.get_definition(labware_id="my-cool-labware-id")
    ).then_return(definition)
    decoy.when(labware_validation.validate_gripper_compatible(definition)).then_return(
        False
    )

    with pytest.raises(
        errors.LabwareMovementNotAllowedError,
        match="Cannot move labware 'My cool labware' with gripper",
    ):
        await subject.execute(data)


async def test_move_labware_with_gripper_raises_on_ot2(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
) -> None:
    """It should raise an error when using a gripper with robot type of OT2."""
    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )
    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.labware.get_definition(labware_id="my-cool-labware-id")
    ).then_return(
        LabwareDefinition2.model_construct(namespace="spacename")  # type: ignore[call-arg]
    )

    decoy.when(state_view.config).then_return(
        Config(robot_type="OT-2 Standard", deck_type=DeckType.OT2_STANDARD)
    )
    with pytest.raises(errors.NotSupportedOnRobotType):
        await subject.execute(data)


async def test_move_labware_raises_when_moving_fixed_trash_labware(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
) -> None:
    """It should raise an error when trying to move a fixed trash."""
    data = MoveLabwareParams(
        labwareId="my-cool-labware-id",
        newLocation=DeckSlotLocation(slotName=DeckSlotName.FIXED_TRASH),
        strategy=LabwareMovementStrategy.USING_GRIPPER,
    )

    definition = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        parameters=Parameters2.model_construct(  # type: ignore[call-arg]
            loadName="My cool labware", quirks=["fixedTrash"]
        ),
    )

    decoy.when(state_view.labware.get(labware_id="my-cool-labware-id")).then_return(
        LoadedLabware(
            id="my-cool-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
            offsetId=None,
        )
    )
    decoy.when(
        state_view.labware.get_definition(labware_id="my-cool-labware-id")
    ).then_return(definition)

    decoy.when(state_view.labware.is_fixed_trash("my-cool-labware-id")).then_return(
        True
    )

    with pytest.raises(
        errors.LabwareMovementNotAllowedError,
        match="Cannot move fixed trash labware 'My cool labware'.",
    ):
        await subject.execute(data)


async def test_labware_raises_when_moved_onto_itself(
    decoy: Decoy,
    subject: MoveLabwareImplementation,
    state_view: StateView,
) -> None:
    """It should raise when the OnLabwareLocation has the same labware ID as the labware being moved."""
    data = MoveLabwareParams(
        labwareId="the-same-labware-id",
        newLocation=OnLabwareLocation(labwareId="a-cool-labware-id"),
        strategy=LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE,
    )

    decoy.when(state_view.labware.get(labware_id="the-same-labware-id")).then_return(
        LoadedLabware(
            id="the-same-labware-id",
            loadName="load-name",
            definitionUri="opentrons-test/load-name/1",
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
            offsetId=None,
        )
    )

    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=OnLabwareLocation(labwareId="a-cool-labware-id"),
        )
    ).then_return(OnLabwareLocation(labwareId="the-same-labware-id"))

    with pytest.raises(
        errors.LabwareMovementNotAllowedError,
        match="Cannot move a labware onto itself.",
    ):
        await subject.execute(data)



from __future__ import annotations

import enum

from typing_extensions import Final


# Implemented with an enum to support type narrowing via `== OFF_DECK`.
class OffDeckType(enum.Enum):
    """The type of the [`OFF_DECK`][opentrons.protocol_api.OFF_DECK] constant.

    Do not use directly, except in type annotations and `isinstance` calls.
    """

    OFF_DECK = "off-deck"
    WASTE_CHUTE = "waste-chute"


OFF_DECK: Final = OffDeckType.OFF_DECK
"""A special location value, indicating that a labware is not currently on the robot's deck.

See [The Off-Deck Location][the-off-deck-location] for details on using `OFF_DECK` with [`ProtocolContext.move_labware()`][opentrons.protocol_api.ProtocolContext.move_labware].
"""
WASTE_CHUTE: Final = OffDeckType.WASTE_CHUTE


class PlungerPositionTypes(enum.Enum):
    PLUNGER_TOP = "top"
    PLUNGER_BOTTOM = "bottom"
    PLUNGER_BLOWOUT = "blow_out"
    PLUNGER_DROPTIP = "drop_tip"


PLUNGER_TOP: Final = PlungerPositionTypes.PLUNGER_TOP
PLUNGER_BOTTOM: Final = PlungerPositionTypes.PLUNGER_BOTTOM
PLUNGER_BLOWOUT: Final = PlungerPositionTypes.PLUNGER_BLOWOUT
PLUNGER_DROPTIP: Final = PlungerPositionTypes.PLUNGER_DROPTIP


class PipetteActionTypes(enum.Enum):
    ASPIRATE_ACTION = "aspirate"
    DISPENSE_ACTION = "dispense"
    BLOWOUT_ACTION = "blowout"


ASPIRATE_ACTION: Final = PipetteActionTypes.ASPIRATE_ACTION
DISPENSE_ACTION: Final = PipetteActionTypes.DISPENSE_ACTION
BLOWOUT_ACTION: Final = PipetteActionTypes.BLOWOUT_ACTION



"""Tests for CommandView.

DEPRECATED: Testing CommandView independently of CommandStore is no longer helpful.
Try to add new tests to test_command_state.py, where they can be tested together,
treating CommandState as a private implementation detail.
"""

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise
from datetime import datetime
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Type, Union

import pytest

from opentrons_shared_data.errors.codes import ErrorCodes

from .command_fixtures import (
    create_failed_command,
    create_queued_command,
    create_running_command,
    create_succeeded_command,
)
from opentrons.protocol_engine import EngineStatus, errors
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.actions import (
    PauseAction,
    PauseSource,
    PlayAction,
    QueueCommandAction,
    StopAction,
)
from opentrons.protocol_engine.actions.actions import ResumeFromRecoveryAction
from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryType
from opentrons.protocol_engine.errors import ErrorOccurrence, ProtocolCommandFailedError
from opentrons.protocol_engine.state.command_history import CommandEntry, CommandHistory
from opentrons.protocol_engine.state.commands import (
    CommandPointer,
    CommandSlice,
    CommandState,
    CommandView,
    QueueStatus,
    RunResult,
    # todo(mm, 2024-10-24): Avoid testing internal implementation details like
    # _RecoveryTargetInfo. See note above about porting to test_command_state.py.
    _RecoveryTargetInfo,
)
from opentrons.protocol_engine.state.update_types import StateUpdate


def _placeholder_error_recovery_policy(*args: object, **kwargs: object) -> Any:
    """A placeholder `ErrorRecoveryPolicy` for tests that don't care about it.

    That should be all the tests in this file, since error recovery was added
    after this file was deprecated.
    """
    raise NotImplementedError()


def get_command_view(  # noqa: C901
    queue_status: QueueStatus = QueueStatus.SETUP,
    run_completed_at: Optional[datetime] = None,
    run_started_at: Optional[datetime] = None,
    is_door_blocking: bool = False,
    run_result: Optional[RunResult] = None,
    running_command_id: Optional[str] = None,
    queued_command_ids: Sequence[str] = (),
    queued_setup_command_ids: Sequence[str] = (),
    queued_fixit_command_ids: Sequence[str] = (),
    run_error: Optional[errors.ErrorOccurrence] = None,
    failed_command: Optional[CommandEntry] = None,
    command_error_recovery_types: Optional[Dict[str, ErrorRecoveryType]] = None,
    recovery_target_command_id: Optional[str] = None,
    finish_error: Optional[errors.ErrorOccurrence] = None,
    commands: Sequence[cmd.Command] = (),
    latest_command_hash: Optional[str] = None,
    has_entered_error_recovery: bool = False,
) -> CommandView:
    """Get a command view test subject."""
    command_history = CommandHistory()

    if running_command_id:
        command_history._set_running_command_id(running_command_id)
    # TODO(tz, 8-21-24): consolidate all quques into 1 and use append_queued_command
    if queued_command_ids:
        for command_id in queued_command_ids:
            command_history._add_to_queue(command_id)
    if queued_setup_command_ids:
        for command_id in queued_setup_command_ids:
            command_history._add_to_setup_queue(command_id)
    if queued_fixit_command_ids:
        for command_id in queued_fixit_command_ids:
            command_history._add_to_fixit_queue(command_id)
    if commands:
        for index, command in enumerate(commands):
            command_history._add(
                command_id=command.id,
                command_entry=CommandEntry(index=index, command=command),
            )

    state = CommandState(
        command_history=command_history,
        queue_status=queue_status,
        run_completed_at=run_completed_at,
        is_door_blocking=is_door_blocking,
        run_result=run_result,
        run_error=run_error,
        finish_error=finish_error,
        failed_command=failed_command,
        command_error_recovery_types=command_error_recovery_types or {},
        recovery_target=(
            _RecoveryTargetInfo(
                command_id=recovery_target_command_id,
                state_update_if_false_positive=StateUpdate(),
            )
            if recovery_target_command_id is not None
            else None
        ),
        run_started_at=run_started_at,
        latest_protocol_command_hash=latest_command_hash,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        has_entered_error_recovery=has_entered_error_recovery,
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    return CommandView(state=state)


def test_get_by_id() -> None:
    """It should get a command by ID from state."""
    command = create_succeeded_command(command_id="command-id")
    subject = get_command_view(commands=[command])

    assert subject.get("command-id") == command


def test_get_command_bad_id() -> None:
    """It should raise if a requested command ID isn't in state."""
    command = create_succeeded_command(command_id="command-id")
    subject = get_command_view(commands=[command])

    with pytest.raises(errors.CommandDoesNotExistError):
        subject.get("asdfghjkl")


def test_get_all() -> None:
    """It should get all the commands from the state."""
    command_1 = create_succeeded_command(command_id="command-id-1")
    command_2 = create_running_command(command_id="command-id-2")
    command_3 = create_queued_command(command_id="command-id-3")

    subject = get_command_view(commands=[command_1, command_2, command_3])

    assert subject.get_all() == [command_1, command_2, command_3]


def test_get_next_to_execute_returns_first_queued() -> None:
    """It should return the next queued command ID."""
    subject = get_command_view(
        queue_status=QueueStatus.RUNNING,
        queued_command_ids=["command-id-1", "command-id-2"],
        queued_fixit_command_ids=["fixit-id-1", "fixit-id-2"],
    )

    assert subject.get_next_to_execute() == "command-id-1"


@pytest.mark.parametrize(
    "queue_status",
    [QueueStatus.SETUP, QueueStatus.RUNNING],
)
def test_get_next_to_execute_prioritizes_setup_command_queue(
    queue_status: QueueStatus,
) -> None:
    """It should prioritize setup command queue over protocol command queue."""
    subject = get_command_view(
        queue_status=queue_status,
        queued_command_ids=["command-id-1", "command-id-2"],
        queued_setup_command_ids=["setup-command-id"],
    )

    assert subject.get_next_to_execute() == "setup-command-id"


@pytest.mark.parametrize(
    "queue_status",
    [QueueStatus.AWAITING_RECOVERY],
)
def test_get_next_to_execute_prioritizes_fixit_command_queue(
    queue_status: QueueStatus,
) -> None:
    """It should prioritize fixit command queue over protocol command queue."""
    subject = get_command_view(
        queue_status=queue_status,
        queued_command_ids=["command-id-1", "command-id-2"],
        queued_setup_command_ids=["setup-command-id"],
        queued_fixit_command_ids=["fixit-1", "fixit-2"],
    )

    assert subject.get_next_to_execute() == "fixit-1"


def test_get_next_to_execute_returns_none_when_no_queued() -> None:
    """It should return None if there are no queued commands."""
    subject = get_command_view(
        queue_status=QueueStatus.RUNNING,
        queued_command_ids=[],
    )

    assert subject.get_next_to_execute() is None


@pytest.mark.parametrize("queue_status", [QueueStatus.SETUP, QueueStatus.PAUSED])
def test_get_next_to_execute_returns_none_if_not_running(
    queue_status: QueueStatus,
) -> None:
    """It should not return protocol commands if the engine is not running."""
    subject = get_command_view(
        queue_status=queue_status,
        queued_setup_command_ids=[],
        queued_command_ids=["command-id-1", "command-id-2"],
    )
    result = subject.get_next_to_execute()

    assert result is None


def test_get_next_to_execute_returns_no_commands_if_paused() -> None:
    """It should not return any type of command if the engine is paused."""
    subject = get_command_view(
        queue_status=QueueStatus.PAUSED,
        queued_setup_command_ids=["setup-id-1", "setup-id-2"],
        queued_command_ids=["command-id-1", "command-id-2"],
        queued_fixit_command_ids=["fixit-id-1", "fixit-id-2"],
    )
    result = subject.get_next_to_execute()

    assert result is None


def test_get_next_to_execute_returns_no_commands_if_awaiting_recovery_no_fixit() -> (
    None
):
    """It should not return any type of command if the engine is awaiting-recovery."""
    subject = get_command_view(
        queue_status=QueueStatus.AWAITING_RECOVERY,
        queued_setup_command_ids=["setup-id-1", "setup-id-2"],
        queued_command_ids=["command-id-1", "command-id-2"],
        queued_fixit_command_ids=[],
    )
    result = subject.get_next_to_execute()

    assert result is None


@pytest.mark.parametrize("run_result", RunResult)
def test_get_next_to_execute_raises_if_stopped(run_result: RunResult) -> None:
    """It should raise if an engine stop has been requested."""
    subject = get_command_view(run_result=run_result)

    with pytest.raises(errors.RunStoppedError):
        subject.get_next_to_execute()


def test_get_is_running_queue() -> None:
    """It should be able to get if the engine is running."""
    subject = get_command_view(queue_status=QueueStatus.PAUSED)
    assert subject.get_is_running() is False

    subject = get_command_view(queue_status=QueueStatus.RUNNING)
    assert subject.get_is_running() is True

    subject = get_command_view(queue_status=QueueStatus.SETUP)
    assert subject.get_is_running() is False


def test_get_command_is_final() -> None:
    """It should be able to tell if a command is complete."""
    completed_command = create_succeeded_command(command_id="completed-command-id")
    failed_command = create_failed_command(command_id="failed-command-id")
    running_command = create_running_command(command_id="running-command-id")
    pending_command = create_queued_command(command_id="queued-command-id")

    subject = get_command_view(
        commands=[completed_command, failed_command, running_command, pending_command]
    )

    assert subject.get_command_is_final("completed-command-id") is True
    assert subject.get_command_is_final("failed-command-id") is True
    assert subject.get_command_is_final("running-command-id") is False
    assert subject.get_command_is_final("queued-command-id") is False


@pytest.mark.parametrize("run_result", RunResult)
def test_get_command_is_final_when_run_has_result(run_result: RunResult) -> None:
    """Queued commands are final when the run will never execute any more commands."""
    completed_command = create_succeeded_command(command_id="completed-command-id")
    failed_command = create_failed_command(command_id="failed-command-id")
    running_command = create_running_command(command_id="running-command-id")
    pending_command = create_queued_command(command_id="queued-command-id")

    subject = get_command_view(
        commands=[completed_command, failed_command, running_command, pending_command],
        run_result=run_result,
    )

    assert subject.get_command_is_final("completed-command-id") is True
    assert subject.get_command_is_final("failed-command-id") is True
    assert subject.get_command_is_final("running-command-id") is False
    assert subject.get_command_is_final("queued-command-id") is True


def test_get_all_commands_final() -> None:
    """It should return True if no commands queued or running."""
    running_command = create_running_command(command_id="running-command-id")

    subject = get_command_view(queued_command_ids=[])
    assert subject.get_all_commands_final() is True

    subject = get_command_view(queued_command_ids=["queued-command-id"])
    assert subject.get_all_commands_final() is False

    subject = get_command_view(
        queued_command_ids=[],
        running_command_id="running-command-id",
        commands=[running_command],
    )
    assert subject.get_all_commands_final() is False


def test_raise_fatal_command_error() -> None:
    """It should raise the fatal command error."""
    completed_command = create_succeeded_command(command_id="command-id-1")
    failed_command = create_failed_command(
        command_id="command-id-2",
        error=errors.ErrorOccurrence(
            id="some-error-id",
            errorType="PrettyBadError",
            createdAt=datetime(year=2021, month=1, day=1),
            detail="Oh no",
            errorCode="4321",
        ),
    )

    subject = get_command_view(
        queued_command_ids=[],
        running_command_id=None,
        failed_command=CommandEntry(index=1, command=failed_command),
        commands=[completed_command, failed_command],
    )

    with pytest.raises(ProtocolCommandFailedError):
        subject.raise_fatal_command_error()


def test_raise_fatal_command_error_tolerates_failed_setup_commands() -> None:
    """It should not call setup command fatal."""
    completed_command = create_succeeded_command(command_id="command-id-1")
    failed_command = create_failed_command(
        command_id="command-id-2",
        intent=cmd.CommandIntent.SETUP,
        error=errors.ErrorOccurrence(
            id="some-error-id",
            errorType="PrettyBadError",
            createdAt=datetime(year=2021, month=1, day=1),
            detail="Oh no",
            errorCode="4321",
        ),
    )

    subject = get_command_view(
        queued_command_ids=[],
        running_command_id=None,
        commands=[completed_command, failed_command],
    )

    subject.raise_fatal_command_error()  # Should not raise.


def test_get_is_stopped() -> None:
    """It should return true if stop requested and no command running."""
    subject = get_command_view(run_completed_at=None)
    assert subject.get_is_stopped() is False

    subject = get_command_view(run_completed_at=datetime(year=2021, day=1, month=1))
    assert subject.get_is_stopped() is True


def test_get_is_started() -> None:
    """It should return true if start requested and no command running."""
    subject = get_command_view(run_started_at=None)
    assert subject.has_been_played() is False

    subject = get_command_view(run_started_at=datetime(year=2021, day=1, month=1))
    assert subject.has_been_played() is True


def test_get_is_terminal() -> None:
    """It should return true if run is in a terminal state."""
    subject = get_command_view(run_result=None)
    assert subject.get_is_terminal() is False

    subject = get_command_view(run_result=RunResult.SUCCEEDED)
    assert subject.get_is_terminal() is True


class ActionAllowedSpec(NamedTuple):
    """Spec data to test CommandView.validate_action_allowed."""

    subject: CommandView
    action: Union[
        PlayAction,
        PauseAction,
        StopAction,
        QueueCommandAction,
        ResumeFromRecoveryAction,
    ]
    expected_error: Optional[Type[Exception]]


action_allowed_specs: List[ActionAllowedSpec] = [
    # play is allowed if the engine is idle
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.SETUP),
        action=PlayAction(requested_at=datetime(year=2021, month=1, day=1)),
        expected_error=None,
    ),
    # play is allowed if engine is idle, even if door is blocking
    ActionAllowedSpec(
        subject=get_command_view(is_door_blocking=True, queue_status=QueueStatus.SETUP),
        action=PlayAction(requested_at=datetime(year=2021, month=1, day=1)),
        expected_error=None,
    ),
    # play is allowed if the engine is paused
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.PAUSED),
        action=PlayAction(requested_at=datetime(year=2021, month=1, day=1)),
        expected_error=None,
    ),
    # pause is allowed if the engine is running
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.RUNNING),
        action=PauseAction(source=PauseSource.CLIENT),
        expected_error=None,
    ),
    # stop is usually allowed
    ActionAllowedSpec(
        subject=get_command_view(),
        action=StopAction(),
        expected_error=None,
    ),
    # queue command is allowed during setup
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.SETUP),
        action=QueueCommandAction(
            request=cmd.HomeCreate(params=cmd.HomeParams()),
            request_hash=None,
            command_id="command-id",
            created_at=datetime(year=2021, month=1, day=1),
        ),
        expected_error=None,
    ),
    # play is disallowed if paused and door is blocking
    ActionAllowedSpec(
        subject=get_command_view(
            is_door_blocking=True, queue_status=QueueStatus.PAUSED
        ),
        action=PlayAction(requested_at=datetime(year=2021, month=1, day=1)),
        expected_error=errors.RobotDoorOpenError,
    ),
    # play is disallowed if stop has been requested
    ActionAllowedSpec(
        subject=get_command_view(run_result=RunResult.STOPPED),
        action=PlayAction(requested_at=datetime(year=2021, month=1, day=1)),
        expected_error=errors.RunStoppedError,
    ),
    # pause is disallowed if stop has been requested
    ActionAllowedSpec(
        subject=get_command_view(run_result=RunResult.STOPPED),
        action=PauseAction(source=PauseSource.CLIENT),
        expected_error=errors.RunStoppedError,
    ),
    # pause is disallowed if engine is not running
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.SETUP),
        action=PauseAction(source=PauseSource.CLIENT),
        expected_error=errors.PauseNotAllowedError,
    ),
    # pause is disallowed if engine is already paused
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.PAUSED),
        action=PauseAction(source=PauseSource.CLIENT),
        expected_error=errors.PauseNotAllowedError,
    ),
    # stop is disallowed if stop has already been requested
    ActionAllowedSpec(
        subject=get_command_view(run_result=RunResult.STOPPED),
        action=StopAction(),
        expected_error=errors.RunStoppedError,
    ),
    # queue command action is disallowed if stop has already been requested
    ActionAllowedSpec(
        subject=get_command_view(run_result=RunResult.STOPPED),
        action=QueueCommandAction(
            request=cmd.HomeCreate(params=cmd.HomeParams()),
            request_hash=None,
            command_id="command-id",
            created_at=datetime(year=2021, month=1, day=1),
        ),
        expected_error=errors.RunStoppedError,
    ),
    # queue setup command is disallowed if paused
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.PAUSED),
        action=QueueCommandAction(
            request=cmd.HomeCreate(
                params=cmd.HomeParams(),
                intent=cmd.CommandIntent.SETUP,
            ),
            request_hash=None,
            command_id="command-id",
            created_at=datetime(year=2021, month=1, day=1),
        ),
        expected_error=errors.SetupCommandNotAllowedError,
    ),
    # queue setup command is disallowed if running
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.RUNNING),
        action=QueueCommandAction(
            request=cmd.HomeCreate(
                params=cmd.HomeParams(),
                intent=cmd.CommandIntent.SETUP,
            ),
            request_hash=None,
            command_id="command-id",
            created_at=datetime(year=2021, month=1, day=1),
        ),
        expected_error=errors.SetupCommandNotAllowedError,
    ),
    # fixit command is disallowed if not in recovery mode
    ActionAllowedSpec(
        subject=get_command_view(queue_status=QueueStatus.RUNNING),
        action=QueueCommandAction(
            request=cmd.HomeCreate(
                params=cmd.HomeParams(),
                intent=cmd.CommandIntent.FIXIT,
            ),
            request_hash=None,
            command_id="command-id",
            created_at=datetime(year=2021, month=1, day=1),
        ),
        expected_error=errors.FixitCommandNotAllowedError,
    ),
    ActionAllowedSpec(
        subject=get_command_view(
            queue_status=QueueStatus.AWAITING_RECOVERY,
            failed_command=CommandEntry(
                index=2,
                command=create_failed_command(
                    command_id="command-id-3",
                    error=ErrorOccurrence(
                        id="error-id",
                        errorType="ProtocolEngineError",
                        createdAt=datetime(year=2022, month=2, day=2),
                        detail="oh no",
                        errorCode=ErrorCodes.GENERAL_ERROR.value.code,
                    ),
                ),
            ),
        ),
        action=QueueCommandAction(
            request=cmd.HomeCreate(
                params=cmd.HomeParams(),
                intent=cmd.CommandIntent.FIXIT,
            ),
            request_hash=None,
            command_id="command-id",
            created_at=datetime(year=2021, month=1, day=1),
        ),
        expected_error=None,
    ),
    # resume from recovery not allowed if fixit commands in queue
    ActionAllowedSpec(
        subject=get_command_view(
            queue_status=QueueStatus.AWAITING_RECOVERY,
            queued_fixit_command_ids=["fixit-id-1", "fixit-id-2"],
            failed_command=CommandEntry(
                index=2,
                command=create_failed_command(
                    command_id="command-id-3",
                    error=ErrorOccurrence(
                        id="error-id",
                        errorType="ProtocolEngineError",
                        createdAt=datetime(year=2022, month=2, day=2),
                        detail="oh no",
                        errorCode=ErrorCodes.GENERAL_ERROR.value.code,
                    ),
                ),
            ),
        ),
        action=ResumeFromRecoveryAction(StateUpdate()),
        expected_error=errors.ResumeFromRecoveryNotAllowedError,
    ),
    ActionAllowedSpec(
        subject=get_command_view(
            queue_status=QueueStatus.AWAITING_RECOVERY_PAUSED, is_door_blocking=True
        ),
        action=QueueCommandAction(
            request=cmd.unsafe.UnsafeUngripLabwareCreate(
                params=cmd.unsafe.UnsafeUngripLabwareParams(),
                intent=cmd.CommandIntent.FIXIT,
            ),
            request_hash=None,
            command_id="command-id",
            created_at=datetime(year=2021, month=1, day=1),
        ),
        expected_error=None,
    ),
]


@pytest.mark.parametrize(ActionAllowedSpec._fields, action_allowed_specs)
def test_validate_action_allowed(
    subject: CommandView,
    action: Union[PlayAction, PauseAction, StopAction],
    expected_error: Optional[Type[Exception]],
) -> None:
    """It should validate allowed play/pause/stop actions."""
    if expected_error is not None:
        expectation: AbstractContextManager[object] = pytest.raises(expected_error)
    else:
        expectation = does_not_raise()

    with expectation:
        result = subject.validate_action_allowed(action)

    if expected_error is None:
        assert result == action


def test_get_errors() -> None:
    """It should be able to pull all ErrorOccurrences from the store."""
    run_error = errors.ErrorOccurrence(
        id="error-1",
        createdAt=datetime(year=2021, month=1, day=1),
        errorType="ReallyBadError",
        detail="things could not get worse",
        errorCode="4321",
    )
    finish_error = errors.ErrorOccurrence(
        id="error-2",
        createdAt=datetime(year=2022, month=2, day=2),
        errorType="EvenWorseError",
        detail="things got worse",
        errorCode="1234",
    )

    no_error_subject = get_command_view()
    assert no_error_subject.get_error() is None

    just_run_error_subject = get_command_view(run_error=run_error)
    assert just_run_error_subject.get_error() == run_error

    just_finish_error_subject = get_command_view(finish_error=finish_error)
    assert just_finish_error_subject.get_error() == finish_error

    both_errors_subject = get_command_view(
        run_error=run_error, finish_error=finish_error
    )
    both_errors_result = both_errors_subject.get_error()
    assert both_errors_result is not None
    assert both_errors_result.wrappedErrors == [run_error, finish_error]


class GetStatusSpec(NamedTuple):
    """Spec data for get_status tests."""

    subject: CommandView
    expected_status: EngineStatus


get_status_specs: List[GetStatusSpec] = [
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.RUNNING,
            running_command_id=None,
            queued_command_ids=[],
        ),
        expected_status=EngineStatus.RUNNING,
    ),
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.PAUSED,
            run_result=RunResult.SUCCEEDED,
            run_completed_at=None,
        ),
        expected_status=EngineStatus.FINISHING,
    ),
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.PAUSED,
            run_result=RunResult.FAILED,
            run_completed_at=None,
        ),
        expected_status=EngineStatus.FINISHING,
    ),
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.PAUSED,
        ),
        expected_status=EngineStatus.PAUSED,
    ),
    GetStatusSpec(
        subject=get_command_view(
            run_result=RunResult.FAILED,
            run_completed_at=datetime(year=2021, day=1, month=1),
        ),
        expected_status=EngineStatus.FAILED,
    ),
    GetStatusSpec(
        subject=get_command_view(
            run_result=RunResult.SUCCEEDED,
            run_completed_at=datetime(year=2021, day=1, month=1),
            finish_error=errors.ErrorOccurrence(
                id="finish-error-id",
                errorType="finish-error-type",
                createdAt=datetime(year=2021, day=1, month=1),
                detail="finish-error-detail",
            ),
        ),
        expected_status=EngineStatus.FAILED,
    ),
    GetStatusSpec(
        subject=get_command_view(
            run_result=RunResult.SUCCEEDED,
            run_completed_at=datetime(year=2021, day=1, month=1),
        ),
        expected_status=EngineStatus.SUCCEEDED,
    ),
    GetStatusSpec(
        subject=get_command_view(
            run_result=RunResult.STOPPED,
            run_completed_at=None,
        ),
        expected_status=EngineStatus.STOP_REQUESTED,
    ),
    GetStatusSpec(
        subject=get_command_view(
            run_result=RunResult.STOPPED,
            run_completed_at=datetime(year=2021, day=1, month=1),
        ),
        expected_status=EngineStatus.STOPPED,
    ),
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.PAUSED,
            is_door_blocking=True,
        ),
        expected_status=EngineStatus.BLOCKED_BY_OPEN_DOOR,
    ),
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.SETUP,
            is_door_blocking=True,
        ),
        expected_status=EngineStatus.IDLE,
    ),
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.PAUSED,
            is_door_blocking=False,
            run_completed_at=datetime(year=2021, day=1, month=1),
        ),
        expected_status=EngineStatus.PAUSED,
    ),
    GetStatusSpec(
        subject=get_command_view(
            queue_status=QueueStatus.SETUP,
            running_command_id="command-id",
            queued_command_ids=["command-id-1"],
            queued_setup_command_ids=["command-id-2"],
        ),
        expected_status=EngineStatus.IDLE,
    ),
]


@pytest.mark.parametrize(GetStatusSpec._fields, get_status_specs)
def test_get_status(subject: CommandView, expected_status: EngineStatus) -> None:
    """It should set a status according to the command queue and running flag."""
    assert subject.get_status() == expected_status


class GetOkayToClearSpec(NamedTuple):
    """Spec data for get_status tests."""

    subject: CommandView
    expected_is_okay: bool


get_okay_to_clear_specs: List[GetOkayToClearSpec] = [
    GetOkayToClearSpec(
        # Protocol not played yet, no commands queued or ran yet
        subject=get_command_view(
            queue_status=QueueStatus.SETUP,
            running_command_id=None,
            queued_command_ids=[],
            queued_setup_command_ids=[],
        ),
        expected_is_okay=True,
    ),
    GetOkayToClearSpec(
        # Protocol commands are queued but not played yet,
        # no setup commands queued or running
        subject=get_command_view(
            queue_status=QueueStatus.SETUP,
            running_command_id=None,
            queued_setup_command_ids=[],
            queued_command_ids=["command-id"],
            commands=[create_queued_command(command_id="command-id")],
        ),
        expected_is_okay=True,
    ),
    GetOkayToClearSpec(
        # Protocol not played yet, setup commands are queued
        subject=get_command_view(
            queue_status=QueueStatus.SETUP,
            running_command_id=None,
            queued_setup_command_ids=["command-id"],
            commands=[create_queued_command(command_id="command-id")],
        ),
        expected_is_okay=False,
    ),
    GetOkayToClearSpec(
        # Protocol is stopped
        subject=get_command_view(
            run_completed_at=datetime(year=2021, day=1, month=1),
        ),
        expected_is_okay=True,
    ),
]


@pytest.mark.parametrize(GetOkayToClearSpec._fields, get_okay_to_clear_specs)
def test_get_okay_to_clear(subject: CommandView, expected_is_okay: bool) -> None:
    """It should report whether an engine is ok to clear."""
    assert subject.get_is_okay_to_clear() is expected_is_okay


def test_get_running_command_id() -> None:
    """It should return the running command ID."""
    running_command = create_running_command(command_id="command-id")

    subject_with_running = get_command_view(
        running_command_id="command-id", commands=[running_command]
    )
    assert subject_with_running.get_running_command_id() == "command-id"

    subject_without_running = get_command_view(running_command_id=None)
    assert subject_without_running.get_running_command_id() is None


def test_get_current() -> None:
    """It should return the "current" command."""
    subject = get_command_view(
        running_command_id=None,
        queued_command_ids=[],
    )
    assert subject.get_current() is None

    command = create_running_command(
        "command-id",
        command_key="command-key",
        created_at=datetime(year=2021, month=1, day=1),
    )
    subject = get_command_view(
        running_command_id="command-id",
        queued_command_ids=[],
        commands=[command],
    )
    assert subject.get_current() == CommandPointer(
        index=0,
        command_id="command-id",
        command_key="command-key",
        created_at=datetime(year=2021, month=1, day=1),
    )

    command_1 = create_succeeded_command(
        "command-id-1",
        command_key="key-1",
        created_at=datetime(year=2021, month=1, day=1),
    )
    command_2 = create_succeeded_command(
        "command-id-2",
        command_key="key-2",
        created_at=datetime(year=2022, month=2, day=2),
    )
    subject = get_command_view(commands=[command_1, command_2])
    subject._state.command_history._set_most_recently_completed_command_id(command_1.id)

    assert subject.get_current() == CommandPointer(
        index=1,
        command_id="command-id-2",
        command_key="key-2",
        created_at=datetime(year=2022, month=2, day=2),
    )

    command_1 = create_succeeded_command(
        "command-id-1",
        command_key="key-1",
        created_at=datetime(year=2021, month=1, day=1),
    )
    command_2 = create_failed_command(
        "command-id-2",
        command_key="key-2",
        created_at=datetime(year=2022, month=2, day=2),
    )
    subject = get_command_view(commands=[command_1, command_2])
    subject._state.command_history._set_most_recently_completed_command_id(command_1.id)

    assert subject.get_current() == CommandPointer(
        index=1,
        command_id="command-id-2",
        command_key="key-2",
        created_at=datetime(year=2022, month=2, day=2),
    )


def test_get_slice_empty() -> None:
    """It should return a slice from the tail if no current command."""
    subject = get_command_view(commands=[])
    result = subject.get_slice(cursor=0, length=2, include_fixit_commands=True)

    assert result == CommandSlice(commands=[], cursor=0, total_length=0)


def test_get_slice() -> None:
    """It should return a slice of all commands."""
    command_1 = create_succeeded_command(command_id="command-id-1")
    command_2 = create_running_command(command_id="command-id-2")
    command_3 = create_queued_command(command_id="command-id-3")
    command_4 = create_queued_command(command_id="command-id-4")

    subject = get_command_view(commands=[command_1, command_2, command_3, command_4])

    result = subject.get_slice(cursor=1, length=3, include_fixit_commands=True)

    assert result == CommandSlice(
        commands=[command_2, command_3, command_4],
        cursor=1,
        total_length=4,
    )

    result = subject.get_slice(cursor=-3, length=10, include_fixit_commands=True)

    assert result == CommandSlice(
        commands=[command_1, command_2, command_3, command_4],
        cursor=0,
        total_length=4,
    )


def test_get_slice_default_cursor_no_current() -> None:
    """It should return a slice from the tail if no current command."""
    command_1 = create_succeeded_command(command_id="command-id-1")
    command_2 = create_succeeded_command(command_id="command-id-2")
    command_3 = create_succeeded_command(command_id="command-id-3")
    command_4 = create_succeeded_command(command_id="command-id-4")

    subject = get_command_view(commands=[command_1, command_2, command_3, command_4])

    result = subject.get_slice(cursor=None, length=3, include_fixit_commands=True)

    assert result == CommandSlice(
        commands=[command_2, command_3, command_4],
        cursor=1,
        total_length=4,
    )


def test_get_slice_default_cursor_failed_command() -> None:
    """It should return a slice from the last executed command."""
    command_1 = create_failed_command(command_id="command-id-1")
    command_2 = create_failed_command(command_id="command-id-2")
    command_3 = create_failed_command(
        command_id="command-id-3",
        error=ErrorOccurrence(
            id="error-id",
            errorType="ProtocolEngineError",
            createdAt=datetime(year=2022, month=2, day=2),
            detail="oh no",
            errorCode=ErrorCodes.GENERAL_ERROR.value.code,
        ),
    )
    command_4 = create_failed_command(command_id="command-id-4")

    subject = get_command_view(
        commands=[command_1, command_2, command_3, command_4],
        run_result=RunResult.FAILED,
        failed_command=CommandEntry(index=2, command=command_3),
    )

    result = subject.get_slice(cursor=None, length=3, include_fixit_commands=True)

    assert result == CommandSlice(
        commands=[command_2, command_3, command_4],
        cursor=1,
        total_length=4,
    )


def test_get_slice_default_cursor_running() -> None:
    """It should select a cursor based on the running command, if present."""
    command_1 = create_succeeded_command(command_id="command-id-1")
    command_2 = create_succeeded_command(command_id="command-id-2")
    command_3 = create_running_command(command_id="command-id-3")
    command_4 = create_queued_command(command_id="command-id-4")
    command_5 = create_queued_command(command_id="command-id-5")

    subject = get_command_view(
        commands=[command_1, command_2, command_3, command_4, command_5],
        running_command_id="command-id-3",
    )

    result = subject.get_slice(cursor=None, length=2, include_fixit_commands=True)

    assert result == CommandSlice(
        commands=[command_2, command_3],
        cursor=1,
        total_length=5,
    )


def test_get_slice_without_fixit() -> None:
    """It should filter out fixit commands when requested."""
    command_1 = create_succeeded_command(command_id="command-id-1")
    command_2 = create_succeeded_command(command_id="command-id-2")
    command_3 = create_running_command(command_id="command-id-3")
    command_4 = create_queued_command(command_id="command-id-4")
    command_5 = create_queued_command(command_id="command-id-5")
    command_6 = create_queued_command(
        command_id="fixit-id-1", intent=cmd.CommandIntent.FIXIT
    )
    command_7 = create_queued_command(
        command_id="fixit-id-2", intent=cmd.CommandIntent.FIXIT
    )

    subject = get_command_view(
        commands=[
            command_1,
            command_2,
            command_3,
            command_4,
            command_5,
            command_6,
            command_7,
        ],
        queued_command_ids=[
            "command-id-1",
            "command-id-2",
            "command-id-3",
            "command-id-4",
            "command-id-5",
            "fixit-id-1",
            "fixit-id-2",
        ],
        queued_fixit_command_ids=["fixit-id-1", "fixit-id-2"],
    )

    result = subject.get_slice(cursor=None, length=7, include_fixit_commands=False)

    assert result == CommandSlice(
        commands=[command_1, command_2, command_3, command_4, command_5],
        cursor=0,
        total_length=5,
    )


def test_get_slice_large_length() -> None:
    """It should handle cases where length is larger than available commands."""
    command_1 = create_succeeded_command(command_id="command-id-1")
    command_2 = create_succeeded_command(command_id="command-id-2")
    command_3 = create_running_command(command_id="command-id-3")

    subject = get_command_view(
        commands=[command_1, command_2, command_3],
        running_command_id="command-id-3",
    )

    result = subject.get_slice(cursor=None, length=10, include_fixit_commands=True)

    assert result == CommandSlice(
        commands=[command_1, command_2, command_3],
        cursor=0,
        total_length=3,
    )


def test_get_slice_explicit_cursor_with_length() -> None:
    """It should use the cursor as the start position when explicitly provided."""
    command_1 = create_succeeded_command(command_id="command-id-1")
    command_2 = create_succeeded_command(command_id="command-id-2")
    command_3 = create_succeeded_command(command_id="command-id-3")
    command_4 = create_succeeded_command(command_id="command-id-4")
    command_5 = create_succeeded_command(command_id="command-id-5")

    subject = get_command_view(
        commands=[command_1, command_2, command_3, command_4, command_5],
    )

    result = subject.get_slice(cursor=1, length=3, include_fixit_commands=True)

    assert result == CommandSlice(
        commands=[command_2, command_3, command_4],
        cursor=1,
        total_length=5,
    )



"""Test thermocycler plate lifter execution side effects."""

from __future__ import annotations

import pytest
from decoy import Decoy, matchers

from opentrons.hardware_control.modules import Thermocycler
from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler
from opentrons.protocol_engine.execution.thermocycler_plate_lifter import (
    ThermocyclerPlateLifter,
)
from opentrons.protocol_engine.state.module_substates import (
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    ModuleLocation,
    ModuleModel,
)
from opentrons.types import DeckSlotName


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore instance."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def equipment(decoy: Decoy) -> EquipmentHandler:
    """Get a mocked out EquipmentHandler instance."""
    return decoy.mock(cls=EquipmentHandler)


@pytest.fixture
def movement(decoy: Decoy) -> MovementHandler:
    """Get a mocked out MovementHandler."""
    return decoy.mock(cls=MovementHandler)


@pytest.fixture
def subject(
    state_store: StateStore,
    equipment: EquipmentHandler,
    movement: MovementHandler,
) -> ThermocyclerPlateLifter:
    """Get ThermocyclerPlateLifter with its dependencies mocked out."""
    return ThermocyclerPlateLifter(
        state_store=state_store,
        equipment=equipment,
        movement=movement,
    )


async def test_lift_plate_for_labware_movement_from_tc_gen2(
    decoy: Decoy,
    state_store: StateStore,
    equipment: EquipmentHandler,
    movement: MovementHandler,
    subject: ThermocyclerPlateLifter,
) -> None:
    """It should execute plate lift if moving labware from TC Gen2."""
    labware_location = ModuleLocation(moduleId="thermocycler-id")
    tc_hardware = decoy.mock(cls=Thermocycler)

    decoy.when(state_store.modules.get_connected_model("thermocycler-id")).then_return(
        ModuleModel.THERMOCYCLER_MODULE_V2
    )
    decoy.when(
        equipment.get_module_hardware_api(
            ThermocyclerModuleId(labware_location.moduleId)
        )
    ).then_return(tc_hardware)
    decoy.when(
        state_store.modules.get_thermocycler_module_substate("thermocycler-id")
    ).then_return(
        ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("abc"),
            is_lid_open=True,
            target_block_temperature=None,
            target_lid_temperature=None,
        )
    )

    async with subject.lift_plate_for_labware_movement(
        labware_location=labware_location
    ):
        decoy.verify(
            await movement.home(axes=None),
            await tc_hardware.lift_plate(),
            await tc_hardware.raise_plate(),
        )
    decoy.verify(
        await tc_hardware.return_from_raise_plate(),
    )


async def test_do_not_lift_plate_if_not_in_tc_gen2(
    decoy: Decoy,
    state_store: StateStore,
    movement: MovementHandler,
    subject: ThermocyclerPlateLifter,
) -> None:
    """It should execute plate lift if moving labware from TC Gen2."""
    decoy.when(state_store.modules.get_connected_model("thermocycler-id")).then_return(
        ModuleModel.THERMOCYCLER_MODULE_V1
    )
    async with subject.lift_plate_for_labware_movement(
        labware_location=ModuleLocation(moduleId="thermocycler-id")
    ):
        pass
    decoy.verify(
        await movement.home(axes=matchers.Anything()),
        times=0,
    )

    async with subject.lift_plate_for_labware_movement(
        labware_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2)
    ):
        pass
    decoy.verify(
        await movement.home(axes=matchers.Anything()),
        times=0,
    )


async def test_do_not_lift_plate_with_lid_closed(
    decoy: Decoy,
    state_store: StateStore,
    equipment: EquipmentHandler,
    movement: MovementHandler,
    subject: ThermocyclerPlateLifter,
) -> None:
    """It should not issue plate lift if lid is not open."""
    labware_location = ModuleLocation(moduleId="thermocycler-id")
    tc_hardware = decoy.mock(cls=Thermocycler)

    decoy.when(state_store.modules.get_connected_model("thermocycler-id")).then_return(
        ModuleModel.THERMOCYCLER_MODULE_V2
    )
    decoy.when(
        equipment.get_module_hardware_api(
            ThermocyclerModuleId(labware_location.moduleId)
        )
    ).then_return(tc_hardware)
    decoy.when(
        state_store.modules.get_thermocycler_module_substate("thermocycler-id")
    ).then_return(
        ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("abc"),
            is_lid_open=False,
            target_block_temperature=None,
            target_lid_temperature=None,
        )
    )

    with pytest.raises(AssertionError):
        async with subject.lift_plate_for_labware_movement(
            labware_location=labware_location
        ):
            pass
